import os
import pickle

import numpy as np
import gmsh

import fiqus.data.DataFiQuSConductor as geom
from fiqus.utils.Utils import GmshUtils
from fiqus.utils.Utils import FilesAndFolders as UFF
from abc import ABC, abstractmethod
from typing import (Dict, List)



### HELPER FUNCTIONS ###

def cylindrical_to_cartesian(rad, angle, height):
    """
    Convert cylindrical coordinates to Cartesian coordinates.

    :return: A list of Cartesian coordinates [x, y, z].
    :rtype: list
    """
    return [rad*np.cos(angle), rad*np.sin(angle), height]

def cartesian_to_cylindrical(x, y, z):
    """
    Convert Cartesian coordinates to cylindrical coordinates.

    :return: A list of cylindrical coordinates [rad, angle, z].
    :rtype: list
    """
    rad = np.sqrt(x**2 + y**2)
    angle = np.arctan2(y, x)
    return [rad, angle, z]

def rotate_vector(vector, angle):
    """
    Rotate a 3D vector in the xy-plane by a given angle.

    :param vector: The 3D vector to rotate. It should be a list or numpy array of three numbers.
    :type vector: list or numpy.ndarray
    :param angle: The angle by which to rotate the vector, in radians.
    :type angle: float

    :return: The rotated vector.
    :rtype: numpy.ndarray
    """
    rotation_matrix = np.array([[np.cos(angle), -np.sin(angle), 0], [np.sin(angle), np.cos(angle), 0], [0, 0, 1]])
    return np.matmul(rotation_matrix, vector)
    
### END HELPER FUNCTIONS ###


### GEOMETRY ###
class Point:
    """
    A class to represent a point in 3D space.

    :cvar points_registry: A list of all points created. This is a class-level attribute.
    :vartype points_registry: list
    :cvar point_snap_tolerance: The tolerance for snapping points to existing points. This is a class-level attribute.
    :vartype point_snap_tolerance: float

    :ivar pos: A numpy array representing the position of the point in 3D space. This is an instance-level attribute.
    :vartype pos: numpy.ndarray
    :ivar tag: A gmsh tag for the point, default is None. This is an instance-level attribute.
    :vartype tag: int, optional
    """
    points_registry = [] # Global registry of points
    point_snap_tolerance = 1e-10 # Tolerance for snapping points to existing points
    def __init__(self, pos : List[float]):
        """
        Constructs the attributes for the point object.

        :param pos: A list of three floats representing the position of the point in 3D space.
        :type pos: List[float]
        """
        self.pos = np.array(pos)
        self.tag = None

    @classmethod
    def create_or_get(cls, pos : any):
        """
        Creates a new point if no point with the given position exists, otherwise returns the existing one to assure unique points.

        :param pos: The position of the point.
        :type pos: list

        :return: A point object.
        :rtype: Point
        """
        new_point = cls(pos)
        if new_point in cls.points_registry:
            return cls.points_registry[cls.points_registry.index(new_point)]
        
        cls.points_registry.append(new_point)
        return new_point
    
    def create_gmsh_instance(self, meshSize : float = 0, tag : int = -1):
        if self.tag == None:
            self.tag = gmsh.model.occ.addPoint(self.pos[0], self.pos[1], self.pos[2], meshSize, tag)
        else:
            pass
    
    def __repr__(self) -> str:
        return f"Point({self.pos})"
    
    def __eq__(self, o: object) -> bool:
        # Two points sharing the same position (within a margin of point_snap_tolerance) should be considered the same point.
        if isinstance(o, Point):
            return np.linalg.norm(np.array(self.pos) - np.array(o.pos)) < self.point_snap_tolerance
        return False
    
    def __hash__(self) -> int:
        return hash(tuple(self.pos))

### CURVE ELEMENTS ###
    
class Curve(ABC):
    """
    Abstract base class for curves in 3D space.

    :cvar curves_registry: A list of all curves created. This is a class-level attribute.
    :vartype curves_registry: list

    :ivar P1: The start point of the curve. This is an instance-level attribute.
    :vartype P1: Point
    :ivar P2: The end point of the curve. This is an instance-level attribute.
    :vartype P2: Point
    :ivar points: A list of points used by the curve. This is an instance-level attribute.
    :vartype points: list
    :ivar tag: A tag for the curve. This is an instance-level attribute.
    :vartype tag: int
    """
    curves_registry = []
    def __init__(self) -> None:
        """
        Constructs the attributes for the curve object.
        """
        self.P1 : Point = None
        self.P2 : Point = None
        self.points = []
        self.tag = None

        # Curve.curves_registry.append(self)

    def set_points(self, *points : Point):
        """
        Sets the points used by the curve.

        :param points: The points to set.
        :type points: Point
        """
        self.points = list(points)
    
    @abstractmethod
    def create_gmsh_instance(self):
        pass
    
    @classmethod
    @abstractmethod
    def create_or_get(cls, *points):
        pass

    @classmethod
    def get_curve_from_tag(cls, tag):
        """
        Returns the curve with the given tag.

        :param tag: The tag of the curve.
        :type tag: int

        :return: The curve with the given tag, or None if no such curve exists.
        :rtype: Curve or None
        """
        for curve in cls.curves_registry:
            if curve.tag == tag:
                return curve
        return None

    
    @classmethod
    def get_closed_loops(cls, curves):
        """
        Returns a list of lists of curves, where each list of curves defines a closed loop.
        The function assumes that no two closed loops share a curve and that chains of curves do not split into multiple chains.

        :param curves: A list of curves.
        :type curves: list

        :return: A list of lists of curves, where each list of curves defines a closed loop.
        :rtype: list
        """
        closed_loops = []

        def get_curve_link(current_curve, curves, curve_link):
            # Recursive function to find a link of connected curves.
            # Curves is a list of curves to search through, not containing the current curve.
            # Curve_link is a list of curves that are connected to the current curve.
            # Curves are removed from 'curves' as they are added to the curve_link.
            # The function returns when no more curves are connected to the current link.
            for curve in curves:
                for point in [current_curve.P1, current_curve.P2]:
                    if point in [curve.P1, curve.P2]:
                        curve_link.append(curve)
                        curves.remove(curve)
                        return get_curve_link(curve, curves, curve_link)
            return curves, curve_link # Return the remaining curves and the curve link

        while len(curves) > 0:
            curve0 = curves[0]
            curves.remove(curve0)
            curves, curve_link = get_curve_link(curve0, curves, [curve0])

            first_curve_points = set([curve_link[0].P1, curve_link[0].P2])
            last_curve_points = set([curve_link[-1].P1, curve_link[-1].P2])
            
            # TODO: Have to add the case where there is only a single curve in the link. This should not be considered a closed loop.
            if len(curve_link)>2 and len(first_curve_points&last_curve_points) == 1: # Check if the last curve in the link is connected to the first curve in the link. If so, the link is a closed loop.
                closed_loops.append(curve_link)
            # If the link only contains two curves, the curves must be connected at both ends to form a closed loop.
            elif len(curve_link) == 2 and len(first_curve_points&last_curve_points) == 2: 
                closed_loops.append(curve_link)
        
        return closed_loops           

class CircleArc(Curve):
    """
    A class to represent a circular arc, subclass of Curve.

    :ivar P1: The first point of the arc. This is an instance-level attribute.
    :vartype P1: Point
    :ivar P2: The second point of the arc. This is an instance-level attribute.
    :vartype P2: Point
    :ivar C: The center of the arc. This is an instance-level attribute.
    :vartype C: Point
    """
    def __init__(self, P1 : any, C : any, P2 : any) -> None:
        """
        Constructs the attributes for the arc object.

        :param P1: The start point of the arc.
        :type P1: list or Point
        :param P2: The end point of the arc.
        :type P2: list or Point
        :param C: The center of the arc.
        :type C: list or Point
        """
        super().__init__()
        self.P1 = P1 if isinstance(P1, Point) else Point.create_or_get(P1)
        self.P2 = P2 if isinstance(P2, Point) else Point.create_or_get(P2)
        self.C = C if isinstance(C, Point) else Point.create_or_get(C)

        self.set_points(self.P1, self.C, self.P2)

        Curve.curves_registry.append(self)

    def create_gmsh_instance(self, tag : int = -1):
        if self.tag == None: # If the curve has not been created yet
            self.P1.create_gmsh_instance()
            self.P2.create_gmsh_instance()
            self.C.create_gmsh_instance()
            self.tag = gmsh.model.occ.addCircleArc(self.P1.tag, self.C.tag, self.P2.tag, tag)

    @classmethod
    def create_or_get(cls, P1 : any, C : any, P2 : any):
        return cls(P1, C, P2)
    
    def __repr__(self) -> str:
        return f"CircleArc({self.P1.pos}, {self.P2.pos})"

class Line(Curve):
    """
    A class to represent a line, subclass of Curve.

    :ivar P1: the start point of the line
    :type P1: list or Point
    :ivar P2: the end point of the line
    :type P2: list or Point

    """
    def __init__(self, P1 : any, P2 : any) -> None:
        """
        Constructs the attributes for the line object.
        P1 and P2 can be either 'Point' objects or lists of three floats representing the position of the point.

        :param P1: the start point of the line
        :type P1: list or Point
        :param P2: the end point of the line
        :type P2: list or Point
        """
        super().__init__()
        self.P1 = P1 if isinstance(P1, Point) else Point.create_or_get(P1)
        self.P2 = P2 if isinstance(P2, Point) else Point.create_or_get(P2)
        
        self.set_points(self.P1, self.P2)
    
    def get_length(self):
        return np.linalg.norm(self.P1.pos - self.P2.pos)
    
    def __repr__(self) -> str:
        return f"Line from {tuple(self.P1.pos)} to {tuple(self.P2.pos)}"
    
    @classmethod
    def create_or_get(cls, P1 : any, P2 : any):
        """
        Creates a new line if it doesn't exist, or returns the existing one.

        :param P1: the first point of the line
        :type P1: Point
        :param P2: the second point of the line
        :type P2: Point

        :return: a line object
        :rtype: Line
        """

        new_line = cls(P1, P2)
        if new_line in cls.curves_registry:
            return cls.curves_registry[cls.curves_registry.index(new_line)]

        cls.curves_registry.append(new_line)
        return new_line
    
    @classmethod
    def is_colinear(cls, line1, line2):
        """
            Checks if two lines are colinear.
        """
        l1 = line1.P2.pos - line1.P1.pos
        l2 = line2.P2.pos - line2.P1.pos
        # Check if the lines are parallel
        if np.linalg.norm(np.cross(l1, l2)) < 1e-10:
            # Check if the lines are colinear
            if np.linalg.norm(np.cross(l1, line2.P1.pos - line1.P1.pos)) < 1e-10:
                return True
        return False

    @classmethod
    def remove_from_registry(cls, lines):
        """
        Removes a list of lines from the registry.

        :param lines: A list of lines to remove.
        :type lines: list
        """
        for line in lines:
            if line in cls.curves_registry:
                cls.curves_registry.remove(line)
    
    def create_gmsh_instance(self, tag : int = -1):
        if self.tag == None:
            self.P1.create_gmsh_instance()
            self.P2.create_gmsh_instance()
            self.tag = gmsh.model.occ.addLine(self.P1.tag, self.P2.tag, tag)

    def __eq__(self, o: object) -> bool:
        # Two Line-entities specified by the same two points should be considered equal lines.
        # Check if the other object is a line
        if isinstance(o, Line): 
            # Check if the lines have the same points
            return (self.P1 == o.P1 and self.P2 == o.P2) or (self.P1 == o.P2 and self.P2 == o.P1)
        return False

    def __hash__(self) -> int:
        return hash(frozenset([self.P1, self.P2])) # Frozenset is hashable and order-independent

class EllipseArc(Curve):
    """
    A class to represent an elliptical arc, subclass of Curve.

    :ivar P1: The start point of the arc. This is an instance-level attribute.
    :vartype P1: Point
    :ivar P2: The end point of the arc. This is an instance-level attribute.
    :vartype P2: Point
    :ivar C: The center point of the arc. This is an instance-level attribute.
    :vartype C: Point
    :ivar M: The major axis point of the arc (point anywhere on the major axis). This is an instance-level attribute.
    :vartype M: Point
    """
    def __init__(self, P_start : any, P_center : any, P_major : any, P_end : any) -> None:
        """
        Initializes an elliptical arc object.

        :param P_start: The start point of the arc. If not a Point instance, attempts to retrieve or create one.
        :type P_start: any
        :param P_end: The end point of the arc. If not a Point instance, attempts to retrieve or create one.
        :type P_end: any
        :param P_center: The center point of the arc. If not a Point instance, attempts to retrieve or create one.
        :type P_center: any
        :param P_major: The major axis point of the arc (point anywhere on the major axis). If not a Point instance, attempts to retrieve or create one.
        :type P_major: any

        :rtype: None
        """
        super().__init__()
        self.P1 = P_start if isinstance(P_start, Point) else Point.create_or_get(P_start) # Start point
        self.P2 = P_end if isinstance(P_end, Point) else Point.create_or_get(P_end) # End point
        self.C = P_center if isinstance(P_center, Point) else Point.create_or_get(P_center) # Center point
        self.M = P_major if isinstance(P_major, Point) else Point.create_or_get(P_major) # Major axis point (point anywhere on the major axis)

        self.set_points(self.P1, self.C, self.M, self.P2)

        Curve.curves_registry.append(self)
        # self.tag = None

    def create_gmsh_instance(self, tag : int = -1):
        if self.tag == None:
            self.P1.create_gmsh_instance()
            self.P2.create_gmsh_instance()
            self.C.create_gmsh_instance()
            self.M.create_gmsh_instance()
            self.tag = gmsh.model.occ.addEllipseArc(self.P1.tag, self.C.tag, self.M.tag, self.P2.tag, tag)

    def __repr__(self) -> str:
        return f"EllipseArc({self.P1.pos}, {self.P2.pos})"

    @classmethod
    def create_or_get(cls, P_start : any, P_center : any, P_major : any, P_end : any):
        return cls(P_start, P_center, P_major, P_end)


### SURFACE ELEMENTS ###
    
class Surface:
    """
    A class to represent a surface in 3D space.

    :cvar surfaces_registry: A class-level attribute that keeps track of all created surfaces.
    :vartype surfaces_registry: list
    :cvar curve_loop_registry: A class-level attribute that keeps track of all curve loops. Each curve loop is stored as a dict with the keys 'curves' and 'tag'. This list is necessary when creating surfaces with holes. The curve-loops of the holes may already have been created when creating the surface of the hole, in which case we get the tag of the existing curve-loop instead of creating a new one.
    :vartype curve_loop_registry: list

    :ivar boundary_curves: A list of Curve objects that define the closed outer boundary of the surface. This is an instance-level attribute.
    :vartype boundary_curves: list
    :ivar inner_boundary_curves: A list of lists of Curve objects. Each list of curves defines the closed boundary of a hole in the surface. This is an instance-level attribute.
    :vartype inner_boundary_curves: list
    :ivar curve_loop_tag: A unique identifier for the curve loop of the outer boundary, initially set to None. This is an instance-level attribute.
    :vartype curve_loop_tag: int
    :ivar inner_curve_loop_tags: A list of unique identifiers for the inner curve loops. Each tag corresponds to the curve-loop of a hole in the surface. This is an instance-level attribute.
    :vartype inner_curve_loop_tags: list
    :ivar surface_tag: A unique identifier for the surface, initially set to None. This is an instance-level attribute.
    :vartype surface_tag: int
    :ivar physical_boundary_tag: A unique identifier for the physical group of the boundary curves, initially set to None. This is an instance-level attribute.
    :vartype physical_boundary_tag: int
    :ivar physical_boundary_name: A name for the physical group of the boundary curves, initially set to None. This is an instance-level attribute.
    :vartype physical_boundary_name: str
    :ivar physical_inner_boundary_tags: A list of unique identifiers for the physical groups of the inner boundary curves. This is an instance-level attribute.
    :vartype physical_inner_boundary_tags: list
    :ivar physical_inner_boundary_names: A list of names for the physical groups of the inner boundary curves. This is an instance-level attribute.
    :vartype physical_inner_boundary_names: list
    :ivar physical_surface_tag: A unique identifier for the physical group of the surface, initially set to None. This is an instance-level attribute.
    :vartype physical_surface_tag: int
    :ivar physical_surface_name: A name for the physical group of the surface, initially set to None. This is an instance-level attribute.
    :vartype physical_surface_name: str
    :ivar material: The material-ID of the surface. Used to specify material properties in the .regions file, initially set to None. This is an instance-level attribute.
    :vartype material: any
    """

    surfaces_registry = [] 
    curve_loop_registry = [] 
    def __init__(self, boundary_curves : List[Curve] = [], inner_boundary_curves : List[List[Curve]] = []):
        """
        Constructs the attributes for the surface object.

        :param boundary_curves: A list of Curve objects that define the outer boundary of the surface. The curves must form a closed loop.
        :type boundary_curves: list[Curve]
        :param inner_boundary_curves: A list of lists of Curve objects. Each list of curves defines the boundary of a hole in the surface.
        :type inner_boundary_curves: list[list[Curve]]
        """
        self.boundary_curves = boundary_curves
        self.inner_boundary_curves = inner_boundary_curves

        self.curve_loop_tag = None
        self.inner_curve_loop_tags : List[int] = []
        self.surface_tag = None 

        self.physical_boundary_tag = None
        self.physical_boundary_name = None
        self.physical_inner_boundary_tags = [] 
        self.physical_inner_boundary_names = []

        self.physical_surface_tag = None 
        self.physical_surface_name = None 

        self.material = None

        Surface.surfaces_registry.append(self)

    def create_gmsh_instance(self):
        if self.surface_tag == None:
            for curve in self.boundary_curves + sum(self.inner_boundary_curves, []):
                curve.create_gmsh_instance()
            
            
            self.curve_loop_tag = self.create_or_get_curve_loop(self.boundary_curves)
            self.inner_curve_loop_tags = [self.create_or_get_curve_loop(curves) for curves in self.inner_boundary_curves]
            self.surface_tag = gmsh.model.occ.add_plane_surface([self.curve_loop_tag] + self.inner_curve_loop_tags)

    def add_physical_boundary(self, name : str = ''):
        self.physical_boundary_tag = gmsh.model.add_physical_group(1, [curve.tag for curve in self.boundary_curves], name = name)
        self.physical_boundary_name = name
    
    def add_physical_surface(self, name : str = ''):
        self.physical_surface_tag = gmsh.model.add_physical_group(2, [self.surface_tag], name = name)
        self.physical_surface_name = name


    @classmethod
    def update_tags(cls, surfaces):
        """
        Updates the tags of model entities of dimension lower than 2.

        When saving the geometry as a .brep file, these tags may be (seemingly arbitrarily) changed. This method ensures that the tags are consistent and correctly assigned.

        Steps:
        1) Find the tags of all outer boundary curves of every surface.
        2) Divide the boundary curves into groups. Each group of curves are either on the boundary of a single surface or on the intersection of the same multiple surfaces.
        3) Update the tags of the curves in the curve groups. The tags are assigned not to their corresponding curve but to any curve in the group.
        4) Update the tags of the points. Points which are in multiple curve-groups are assigned to a new group. Point tags are assigned based on the groups they are in.

        :param surfaces: A list of Surface objects to update the tags for.
        :type surfaces: list[Surface]
        """

        gmsh.model.occ.synchronize()
        #1) Find the tags of all outer boundary curves of every surface
        #1.1) Find the inner surfaces of each surface with holes
        surfaces_inner_surface_indices = [] # List of lists of indices of surfaces. Each list of indices corresponds to the surfaces that are inside the surface with the same index.
        for surface in surfaces:
            inner_surface_indices = []
            if surface.inner_boundary_curves: # If the surface has holes
                for inner_boundary in surface.inner_boundary_curves: 
                    for surface in surfaces: # Loop through all surfaces to find the surfaces that are inside the outer surface.
                        if set(inner_boundary) & set(surface.boundary_curves): # If the two sets of curves have any common curves, the inner surface is inside the outer surface.
                            inner_surface_indices.append(surfaces.index(surface)) # Add the index of the inner surface to the list of inner surfaces.
            surfaces_inner_surface_indices.append(inner_surface_indices)

        #1.2) Find the tags of the outer boundary curves of each surface by finding the boundary of the surface (inner and outer) and removing the boundary curves of the inner surfaces.
        surfaces_outer_boundary_tags = [set([abs(tag) for dim, tag in gmsh.model.get_boundary([(2, surface.surface_tag)])]) for surface in surfaces]
        for surface_i in range(len(surfaces)):
            if surfaces[surface_i].inner_boundary_curves:
                surface_inner_surfaces_boundary_tags = set.union(*[surfaces_outer_boundary_tags[i] for i in surfaces_inner_surface_indices[surface_i]]) # The boundary tags of the inner surfaces
                surfaces_outer_boundary_tags[surface_i] = surfaces_outer_boundary_tags[surface_i] - surface_inner_surfaces_boundary_tags 

        #2) Divide the boundary curves into groups. Each group of curves are either on the boundary of a single surface or on the intersection of multiple surfaces, but not both.
        # We also find the tags of the curves in each group.
        surface_outer_boundary_curves = [set(surface.boundary_curves) for surface in surfaces] 

        curve_groups = [] # List of sets of curves. Each set of curves is a group, defined as curves which lie only on the boundary of a single surface or on the intersection of multiple surfaces.
        curve_group_tags = [] # List of sets of tags. Each set of tags corresponds to the set of curves in curve_groups.
        for i, si_curves in enumerate(surface_outer_boundary_curves):
            for j, sj_curves in enumerate(surface_outer_boundary_curves[i+1:]):
                j += i+1
                common_curves = si_curves & sj_curves
                if common_curves:
                    curve_groups.append(common_curves) # Add the common curves to the list of curve groups
                    si_curves -= common_curves # Remove the common curves from the surface boundary curves
                    sj_curves -= common_curves # Remove the common curves from the surface boundary curves

                    curve_group_tags.append(surfaces_outer_boundary_tags[i] & surfaces_outer_boundary_tags[j]) # Add the tags of the common curves to the list of curve group tags
                    surfaces_outer_boundary_tags[i] -= curve_group_tags[-1] # Remove the tags of the common curves from the surface boundary tags
                    surfaces_outer_boundary_tags[j] -= curve_group_tags[-1] # Remove the tags of the common curves from the surface boundary tags
 
            curve_groups.append(si_curves) # Add the remaining curves to the list of curve groups
            curve_group_tags.append(surfaces_outer_boundary_tags[i]) # Add the tags of the remaining curves to the list of curve group tags

        #3) Update the tags of the curves in the curve groups
        # The tags are assigned not to their corresponding curve but to any curve in the group.
        for group, group_tags in zip(curve_groups, curve_group_tags):
            for curve, tag in zip(group, group_tags):
                curve.tag = tag
        
        #4) We have now updated the tags of all curves. Next we update the tags of the points.
        
        #4.1) Get all points in each group of curves and the tags of the points in each group of curves
        curve_groups_points = [set([point for curve in group for point in [curve.P1, curve.P2]]) for group in curve_groups]
        curve_groups_point_tags = [set(sum([list(gmsh.model.get_adjacencies(1, tag)[1]) for tag in group_tags], [])) for group_tags in curve_group_tags]

        #4.2) Points which are in multiple curve-groups are assigned to a new group.
        # These points will be removed from the groups they are in and assigned to the new group, based on the groups they are in.
        # Iterate trough all points and check which groups they are in. Points in same groups will be assigned to a new group.
        all_points = set.union(*curve_groups_points)
        point_new_groups = {} # Dictionary with keys as tuples of indices of the groups the point is in, and values as lists of points in the same groups.
        for point in all_points:
            groups_point_is_in = [i for i, group in enumerate(curve_groups_points) if point in group]
            # If the point is in multiple groups, remove the point from all groups
            if len(groups_point_is_in) > 1:
                for i in groups_point_is_in: 
                    curve_groups_points[i].remove(point) # Remove the point from all groups, as it will be assigned to a new group.
                # Sort the groups the point is in, make it a tuple and use it as a key in a dictionary. The value is a list of all points in the same groups.
                point_new_groups[tuple(sorted(groups_point_is_in))] = point_new_groups.get(tuple(sorted(groups_point_is_in)), []) + [point]
        
        #4.3) Update the tags of the points in the new groups
        # Get the tags of all points in each group of points as the boundary of the group of curves
        for group_indices, points in point_new_groups.items():
            # The tags of the points in the new group is the intersection of the tags of the points in the groups the point is in.
            point_tags = set.intersection(*[curve_groups_point_tags[i] for i in group_indices])

            # Update the tags of the points in the group
            for point, point_tag in zip(points, point_tags):
                point.tag = point_tag
            
            # Remove the tags of the points in the new group from the tags of the points in the groups the point was in before it was assigned to the new group.
            for i in group_indices:
                curve_groups_point_tags[i] -= point_tags
        
        #4.4) Update the tags of points in the remaining groups 'curve_groups_points'
        for group, group_tags in zip(curve_groups_points, curve_groups_point_tags):
            for point, tag in zip(group, group_tags):
                point.tag = tag

            


    @classmethod
    def remove_from_registry(cls, surfaces):
        for surface in surfaces:
            cls.surfaces_registry.remove(surface)

    @classmethod
    def create_or_get_curve_loop(cls, curves):
        """
        Creates a curve loop if it does not already exist, otherwise returns the existing curve loop.

        A curve loop is a sequence of curves that form a closed loop. This method checks if a curve loop with the given curves already exists in the curve_loop_registry. 
        If it does, the method returns the tag of the existing curve loop. 
        If it doesn't, the method creates a new curve loop, adds it to the curve_loop_registry, and returns its tag.

        :param curves: A list of Curve objects that define the curve loop.
        :type curves: list[Curve]

        :return: The tag of the curve loop.
        :rtype: int
        """
        for curve_loop in cls.curve_loop_registry:
            if curve_loop['curves'] == set(curves):
                return curve_loop['tag']
        
        tag = gmsh.model.occ.addCurveLoop([curve.tag for curve in curves])
        cls.curve_loop_registry.append({'tag': tag, 'curves': set(curves)})
        return tag
    
    @classmethod
    def replace_overlapping_edges(cls):
        """
        Replaces overlapping boundary-curves of type 'Line' across all surfaces.

        If multiple surface boundaries contain lines which are overlapping, the overlapping lines are replaced by the fraction
        of the line which is unique to the surface as well as the fraction of the line which is shared with the other surface.

        Steps:
        1) Sort all existing lines into groups of lines which are colinear.
        2) For each group of colinear lines, make a list of all the unique points used, sorted by their location on the colinear line.
        3) For each line of each surface, replace the line by fragments of the line, defined by the points in the list from step 2.
        """
        #1) Sort all existing lines into groups of lines which are colinear
        lines = [line for line in Curve.curves_registry if type(line) == Line]
        if len(lines) == 0: # If there are no lines, skip the rest of the function
            return
        colinear_groups = [[lines[0]]]
        for line in lines[1:]:
            for group in colinear_groups:
                if Line.is_colinear(line, group[0]):
                    group.append(line)
                    break
            else:
                colinear_groups.append([line])

        #2) For each group of colinear lines, make a list of all the unique points used, sorted by their location on the colinear line
        colinear_groups_points = []
        for group in colinear_groups:
            points = []
            for line in group:
                points.append(line.P1)
                points.append(line.P2)
            points = list(set(points)) # Remove duplicates
            angle = np.arctan2((line.P2.pos-line.P1.pos)[1], (line.P2.pos-line.P1.pos)[0]) # Angle of the lines with respect to the x-axis
            positions = [p.pos - points[0].pos for p in points] # Move the points so that the colinear line passes through the origin
            positions = [rotate_vector(p, -angle)[0] for p in positions] # Rotate the points so that the lines are parallel to the x-axis, making it a 1D problem.
            points = np.array(points)[np.argsort(positions)].tolist() # Sort the points by their position along the colinear line
            colinear_groups_points.append(points)    

        #3) For each line for each surface, replace the line by fragments of the line, defined by the points in the list from step 2	
        for area in cls.surfaces_registry: # For each rectangle
            for l, line in enumerate(area.boundary_curves): # For each line
                for i, group in enumerate(colinear_groups): # Find the group of colinear lines that the line belongs to
                    if line in group:
                        if len(group) == 1: # If the line is not colinear with any other line, skip it
                            break
                        # Find the points that define the line fragments
                        points = colinear_groups_points[i]
                        line_point1_index = points.index(line.P1)	
                        line_point2_index = points.index(line.P2)
                        if line_point1_index > line_point2_index: 
                            # If the points orientation differs from the line orientation, reverse the list of points
                            # The points are sorted by their position along the colinear line, which is not necessarily the same as the orientation of the line.
                            # The orientation of the fragments must match the orientation of the line.
                            points.reverse()
                            line_point1_index = points.index(line.P1)
                            line_point2_index = points.index(line.P2)
                        line_fragment_points = points[line_point1_index:line_point2_index+1] # The points that define the line fragments

                        # Create the line fragments
                        line_fragments = [Line.create_or_get(line_fragment_points[i], line_fragment_points[i+1]) for i in range(len(line_fragment_points)-1)]
                        
                        if len(line_fragments) > 1: # If the line is split into multiple fragments, remove the old line and insert the new fragments into the boundary curves of the surface.
                            Line.remove_from_registry([line]) # Remove the old line from the lines-registry.
                            for fragment in line_fragments:
                                area.boundary_curves.insert(area.boundary_curves.index(line), fragment) # Insert the new line fragments into the boundary curves of the surface.
                            area.boundary_curves.remove(line) # Remove the original line from the boundary curves of the surface.

    @classmethod
    def set_correct_boundary_orientation(self, surfaces):
        """
        When creating surfaces with holes, the boundaries must have a specific orientation to ensure that the surfaces are correctly defined.
        This method sets the correct orientation of the boundaries of the surfaces.
        The orientation of the boundary is determined by the orientation of the curve-loop, which is determined by the orientation of the first curve in the loop.
        We therefore must ensure that the first curve of the curve-loop is oriented correctly.
        """	
        # It seems to work to just set all the boundaries to have the same orientation. Very simple and does not require any complex logic.
        for surface in surfaces:
            outer_boundary_points = sum([[curve.P1, curve.P2] for curve in surface.boundary_curves], [])
            mean_point = np.mean([point.pos for point in outer_boundary_points], axis=0) # Mean point of the boundary (Center of mass)
            mean_to_P1 = surface.boundary_curves[0].P1.pos - mean_point # Vector from the mean point to the first point of the boundary curve
            mean_to_P2 = surface.boundary_curves[0].P2.pos - mean_point # Vector from the mean point to the second point of the boundary curve

            # Using the two vectors we can check the orientation of the first curve in the loop, based on the sign of the determinant of the matrix formed by the two vectors.
            if np.linalg.det(np.column_stack((mean_to_P1[:-1], mean_to_P2[:-1]))) < 0:
                # If the determinant is negative we reverse the orientation so that it is positive (counter-clockwise orientation)
                surface.boundary_curves[0].P1, surface.boundary_curves[0].P2 = surface.boundary_curves[0].P2, surface.boundary_curves[0].P1

class Disk(Surface):
    """
    A class to represent a disk. Inherits from the Surface class.

    :ivar rad: The radius of the disk. This is an instance-level attribute.
    :vartype rad: float
    :ivar partitions: The number of partitions to divide the boundary of the disk into when generating boundary curves. This is an instance-level attribute.
    :vartype partitions: int
    :ivar center_point: The center point of the disk. Can be a position ([x, y, z]) or a 'Point' object. This is an instance-level attribute.
    :vartype center_point: Point
    :ivar physicalEdgePointTag: A unique identifier for the physical group of the edge point. Used in pro-template for fixing phi=0 in the outer matrix boundary as well as on the filament boundaries. This is an instance-level attribute.
    :vartype physicalEdgePointTag: int
    """
    def __init__(self, center_point : any, rad : float, partitions : int = 4):
        """
        Constructs the attributes for the disk object.

        :param center_point: The center point of the disk. Can be a position ([x, y, z]) or a 'Point' object.
        :type center_point: any
        :param rad: The radius of the disk.
        :type rad: float
        :param partitions: The number of partitions to divide the boundary of the disk into when generating boundary curves, defaults to 4
        :type partitions: int, optional
        """
        super().__init__()
        self.rad = rad
        self.partitions = partitions

        self.center_point = center_point if isinstance(center_point, Point) else Point.create_or_get(center_point)
        self.boundary_curves = self.generate_boundary_curves()
        self.inner_boundary_curves = []

        self.physicalEdgePointTag = None # Used in pro-template for fixing phi=0 in the outer matrix boundary as well as on the filament boundaries.

    def generate_boundary_curves(self):
        """
        Generates the boundary curves for the disk.

        This method divides the boundary of the disk into 'partitions' number of segments and creates a CircleArc for each segment.

        :return: A list of CircleArc objects that define the boundary of the disk.
        :rtype: list
        """
        edgePoints = [np.array(self.center_point.pos) + np.array(cylindrical_to_cartesian(self.rad, 2*np.pi * n/self.partitions + np.pi + cartesian_to_cylindrical(self.center_point.pos[0], self.center_point.pos[1], self.center_point.pos[2])[1], 0)) for n in range(self.partitions)]
        boundary_curves = [CircleArc(edgePoints[n], self.center_point, edgePoints[(n+1)%self.partitions]) for n in range(self.partitions)]
        return boundary_curves
    
class Hexagon(Surface):
    """
    A class to represent a hexagon. Inherits from the Surface class.

    :ivar rad: The radius of the hexagon.
    :vartype rad: float
    :ivar center_point: The center point of the hexagon. Can be a position ([x, y, z]) or a 'Point' object.
    :vartype center_point: Point
    :ivar rotation: The rotation of the hexagon in radians.
    :vartype rotation: float
    """
    def __init__(self, center_point : any, rad : float, rotation : float = 0):
        """
        Constructs the attributes for the hexagon object.

        :param center_point: The center point of the hexagon. Can be a position ([x, y, z]) or a 'Point' object.
        :type center_point: any
        :param rad: The radius of the hexagon.
        :type rad: float
        :param rotation: The rotation of the hexagon in radians.
        :type rotation: float
        """
        super().__init__()
        self.rad = rad
        self.center_point = center_point if isinstance(center_point, Point) else Point.create_or_get(center_point)
        self.rotation = rotation
        self.boundary_curves = self.generate_boundary_curves()
        self.inner_boundary_curves = []

    def generate_boundary_curves(self):
        """
        Generates the boundary curves for the hexagon.

        This method creates the boundary of the hexagon by rotating a vector from the center-point to the first edge-point.

        :return: A list of Line objects that define the boundary of the hexagon.
        :rtype: list
        """
        edgePoints = [np.array(self.center_point.pos) + rotate_vector(np.array([self.rad, 0, 0]), 2*np.pi * n/6 + self.rotation) for n in range(6)]
        boundary_curves = [Line.create_or_get(edgePoints[n], edgePoints[(n+1)%6]) for n in range(6)]
        return boundary_curves

### END GEOMETRY ###


class TwistedStrand:
    """
    A class to represent a 2D cross section of a twisted strand.

    :ivar filaments: Each list of surfaces represent six filaments in the same layer.
    :vartype filaments: list[list[Surface]]
    :ivar Matrix: List of surfaces corresponding to the matrix partitions.
    :vartype Matrix: list[Surface]
    :ivar Air: A surface representing the air region.
    :vartype Air: Surface
    """
    def __init__(self) -> None:
        """
        Initializes the TwistedStrand object.
        """
        self.filaments : List[List[Surface]] = []
        self.filament_holes : List[List[Surface]] = [] # If the filaments have holes (only for certain geometries obtained via YAML files)
        self.matrix : List[Surface] = [] # Inner, middle, outer 
        self.air : Surface = None

    def create_filament_center_points(self, N_points : int, filament_radius : float, outer_boundary_radius : float, inner_boundary_radius : float, circular_filament_distribution = True):
        """
        Creates the center-points of N_points filaments. The points are distributed in layers---the first layer containing 6 points, the second 12, the third 18, etc---and groups of
        6 points satisfying rotational symmetry when rotated by pi/3. The first layer is composed of one group, the second of two groups, etc.

        :param N_points: The number of points to create.
        :type N_points: int
        :param filament_radius: The radius of the filament.
        :type filament_radius: float
        :param outer_boundary_radius: The radius of the outer boundary.
        :type outer_boundary_radius: float
        :param inner_boundary_radius: The radius of the inner boundary.
        :type inner_boundary_radius: float
        :param circular_filament_distribution: If True, points are distributed in a circular pattern. If False, points are distributed in a hexagonal pattern. Defaults to True.
        :type circular_filament_distribution: bool, optional

        :return: A list of lists of lists of points representing the filament center-points for each filament in each group. Each point is a list of three coordinates [x, y, z] and each list of points represents a group of 6 points satisfying rotational symmetry.
        :rtype: list[list[list[float]]]
        """
        # The hexagonal grid can be created by transforming the cartesian coordinate grid to a hexagonal grid. 
        # Point centers are created as unit vectors in the hexagonal grid, and then transformed to cartesian coordinates.
        # This function takes as input a vector in the hexagonal grid and returns its distance from the origin in the cartesian coordinate system.
        point_relative_magnitude = lambda v : np.sqrt(v[0]**2 + 2*v[0]*v[1]*np.cos(np.pi/3) + v[1]**2) 
        def create_first_hexant_points(N_points_per_hexant, outer_inner_boundary_ratio, possible_next_points = np.array([[1,0]]), points = np.empty((0,2))):
            """
            Recursive algorithm used to create the first hexant of an hexagonal point grid. These points can subsequently be rotated by n*pi/3 (for n in [1,5]) radians to create the next points.
            The function returns the points in the first sextant of the matrix, sorted by distance from the origin.

            The points are created in an iterative manner, starting from the points closest to the origin and moving outwards. 
            Selecting the next point is done either by selecting the point with the smallest distance from the origin (giving a circular distribution), or by prioritizing points in the 
            same layer as the previous point, choosing the point closest to the 30 degree angle (giving an hexagonal distribution). If no points are available in the same layer, the next layer is considered.
            Choosing points by distance from the origin results in a more circular distribution of points, while choosing points by angle results in a hexagonal distribution of points.
            """

            if circular_filament_distribution:
                # The next point is the one with the smallest distance from the origin
                next_point = min(possible_next_points, key=point_relative_magnitude)
            
            else: 
                #1) Filter possible next points to include only points in the lowest not full layer.
                next_points_in_same_layer = possible_next_points[possible_next_points.sum(axis=1) == possible_next_points.sum(axis=1).min()]
                #2) Choose the next point as the one with an angle closest to 30 degrees.
                next_point = min(next_points_in_same_layer, key=lambda v : np.abs(v[1]-v[0]))

            possible_next_points = np.delete(possible_next_points, np.where(np.all(possible_next_points == next_point, axis=1)), axis=0) # Remove the selected point from possible_next_points
            points = np.append(points, np.array([next_point]), axis=0) # Add the selected point to points
            points = sorted(points, key=lambda p: point_relative_magnitude(p)) # Sort the points by their distance from the origin

            new_possible_next_points = np.array([next_point + np.array([1,0]), next_point + np.array([0,1])]) # The possible next points from the current point
            possible_next_points = np.unique( np.concatenate( (new_possible_next_points, possible_next_points) ), axis = 0 ) # Add the new possible next points to the list of possible next points
            

            if len(points) == N_points_per_hexant:
                # Check if the outer-inner boundary ratio is satisfied.
                # The outermost points are always placed at the outer boundary, if the outer-inner boundary ratio is not satisfied, the innermost points must be within the inner boundary.
                # The innermost point is then removed and replaced by a point in the outermost layer.
                if point_relative_magnitude(points[-1]) / point_relative_magnitude(points[0]) <= outer_inner_boundary_ratio:
                    return points
                else:
                    N_points_per_hexant += 1
                    points = np.delete(points, 0, axis=0)
                    return create_first_hexant_points(N_points_per_hexant, outer_inner_boundary_ratio, possible_next_points, points)

            else:
                return create_first_hexant_points(N_points_per_hexant, outer_inner_boundary_ratio, possible_next_points, points)
            
        outer_boundary_radius -= filament_radius*1.1
        if inner_boundary_radius != 0:
            inner_boundary_radius += filament_radius*1.1 

        outer_inner_boundary_ratio = outer_boundary_radius / inner_boundary_radius if inner_boundary_radius != 0 else 1e10
        if N_points % 6 == 1 and inner_boundary_radius == 0:
            N_points -= 1
            points = [[[0,0,0]]]
        else:
            points = []

        groups_of_rotational_symmetry_first_points = create_first_hexant_points(N_points//6, outer_inner_boundary_ratio) # The first hexant of points, sorted by distance from the origin

        R = outer_boundary_radius / point_relative_magnitude(groups_of_rotational_symmetry_first_points[-1]) # Scaling factor to place points at the correct distance from the origin
        cart_to_hex_transformation = np.array([[R, R*np.cos(np.pi/3)], [0, R*np.sin(np.pi/3)]]) # Transformation matrix from cartesian to hexagonal coordinates

        # Rotate the points to create the other hexants of points
        for point in groups_of_rotational_symmetry_first_points: 
            transformed_point = np.matmul(cart_to_hex_transformation, point) # Transform the point from the cartesian coordinate system to the hexagonal coordinate system
            point_group = []
            rotation_angle = 0
            for hexagon_side in range(0, 6):
                rotation_matrix = np.array([[np.cos(rotation_angle), -np.sin(rotation_angle)], [np.sin(rotation_angle), np.cos(rotation_angle)]])

                rotated_point = np.matmul(rotation_matrix, transformed_point)
                point_group.append(list(rotated_point)+[0])

                rotation_angle += np.pi/3
            points.append(point_group)
        return points
    
    def create_geometry(self, filament_radius : float, hexagonal_filaments : bool, N_filaments : int, matrix_inner_radius : float, matrix_middle_radius : float, matrix_outer_radius : float, circular_filament_distribution : bool = False):
        """
        Creates the full geometry of the strand cross-section.

        This method generates the geometry by creating a grid of points that represent the center points of the filaments. 
        It then creates the filaments at these points. The filaments can be either hexagonal or circular, depending on the `hexagonal_filaments` parameter. 
        Finally, it creates the matrix that contains the filaments. The matrix is divided into an inner, middle, and outer area. The middle section contains the filaments, and the inner and outer areas are 'empty'.

        :param filament_radius: The radius of the filaments.
        :type filament_radius: float
        :param hexagonal_filaments: If True, the filaments are hexagonal. If False, the filaments are circular.
        :type hexagonal_filaments: bool
        :param N_filaments: The number of filaments to create.
        :type N_filaments: int
        :param matrix_inner_radius: The radius of the inner area of the matrix.
        :type matrix_inner_radius: float
        :param matrix_middle_radius: The radius of the middle area of the matrix. This is where the filaments are located.
        :type matrix_middle_radius: float
        :param matrix_outer_radius: The radius of the outer area of the matrix.
        :type matrix_outer_radius: float
        :param circular_filament_distribution: If True, the filaments are distributed in a circular pattern. If False, the filaments are distributed in a hexagonal pattern. Defaults to False.
        :type circular_filament_distribution: bool, optional
        """
        # 1) Create the point-grid representing the center points of the filaments
        # 1.1) Get the point positions
        filament_centers = self.create_filament_center_points(N_filaments, filament_radius, matrix_middle_radius, matrix_inner_radius, circular_filament_distribution)
        # 1.2) Create the center points
        center_points = []
        for layer in filament_centers:
            layer_points = []
            for pos in layer:
                P = Point.create_or_get(pos)
                layer_points.append(P)
            center_points.append(layer_points)

        # 2) Create the filaments 
        filaments = []
        for layer_n in range(len(center_points)):
            layer_filaments = []
            for point_i in range(len(center_points[layer_n])):
                if hexagonal_filaments:
                    filament = Hexagon(center_points[layer_n][point_i], filament_radius, np.pi/6) 
                else:
                    filament = Disk(center_points[layer_n][point_i], filament_radius)
                layer_filaments.append(filament)
            filaments.append(layer_filaments)
        self.filaments = filaments

        # 3) Create the matrix
        # The matrix will be divided into an inner-, middle- and outer area. The middle section contains the filaments and the inner and outer areas are 'empty'.
        # No inner section will be made if the matrix inner_radius is 0.
        if matrix_inner_radius != 0: 
            inner_section = Disk([0,0,0], matrix_inner_radius)
            middle_section = Disk([0,0,0], matrix_middle_radius) # Middle section
            middle_section.inner_boundary_curves.append(inner_section.boundary_curves)
            for layer in self.filaments:
                for filament in layer:
                    middle_section.inner_boundary_curves.append(filament.boundary_curves)
        else:
            middle_section = Disk([0,0,0], matrix_middle_radius) # Middle section
            for layer in self.filaments:
                for filament in layer:
                    middle_section.inner_boundary_curves.append(filament.boundary_curves)

        outer_section = Disk([0,0,0], matrix_outer_radius)
        outer_section.inner_boundary_curves.append(middle_section.boundary_curves)
        
        self.matrix = [middle_section, outer_section] if matrix_inner_radius == 0 else [inner_section, middle_section, outer_section]

    def add_air(self, rad : float):
        # The air region is defined as the region from the matrix outer boundary to the radius 'rad'. The air radius must be greater than the matrix radius.
        self.air = Disk([0,0,0], rad) # Initialize the air region as a disk with radius 'rad'

        # Find the combined outer boundary of the strand geometry, which is the inner boundary of the air region.
        # The outer boundary of the strand geometry is is not necessarily the outer boundary of the matrix, as the outer matrix partition may not fully contain the full strand (as with a wire-in-channel geometry).
        strand_outer_boundary = set(self.matrix[0].boundary_curves) # Start with the boundary of the inner matrix partition
        for i, matrix_partition in enumerate(self.matrix[:-1]): # Loop over the matrix partitions
            next_matrix_partition = self.matrix[i+1]
            
            inner_partition_boundary = set(matrix_partition.boundary_curves)
            next_partition_boundary = set(next_matrix_partition.boundary_curves)

            if inner_partition_boundary & next_partition_boundary: # If the inner and outer partition boundaries share some curves
                strand_outer_boundary = strand_outer_boundary ^ next_partition_boundary # Get the combined boundary of the inner and outer partition boundaries
            else:
                strand_outer_boundary = next_partition_boundary # If the inner and outer partition boundaries do not share any curves, the outer boundary is simply the boundary of the outer partition.

        strand_outer_boundary = Curve.get_closed_loops(list(strand_outer_boundary))[0] # Simply used to sort the curves in the outer boundary into a correct order which can be used to create a closed loop.

        self.air.inner_boundary_curves.append(strand_outer_boundary)

    def update_tags(self):
        """
        When the geometry is loaded from a .brep file, the tags of entities with dimensions lower than the highest dimension are not preserved and may change unpredictably.
        This function updates the tags of the points, curves and surfaces in the geometry to ensure that they are consistent with the current gmsh model.
        """
        surfaces = sum(self.filaments, []) + sum(self.filament_holes, []) + self.matrix + [self.air] 
        Surface.update_tags(surfaces)

    def create_gmsh_instance(self):
        """
            Creates the gmsh instances of the geometry.
        """
        surfaces = sum(self.filaments, []) + sum(self.filament_holes, []) + self.matrix + [self.air]

        Surface.set_correct_boundary_orientation(surfaces)

        for surface in surfaces:
            surface.create_gmsh_instance()
        
    def add_physical_groups(self):
        """
            Creates all physical groups.
        """
        # Filaments: Add physical boundary and surface
        for layer_n in range(len(self.filaments)):
            for filament_i in range(len(self.filaments[layer_n])):
                self.filaments[layer_n][filament_i].add_physical_boundary(f"Boundary: Filament {filament_i} in layer {layer_n}")
                self.filaments[layer_n][filament_i].add_physical_surface(f"Surface: Filament {filament_i} in layer {layer_n}")

                self.filaments[layer_n][filament_i].physicalEdgePointTag = gmsh.model.addPhysicalGroup(0, [self.filaments[layer_n][filament_i].boundary_curves[0].points[0].tag])

        # Add physical surface for the filament holes
        for layer_n in range(len(self.filament_holes)):
            for filament_i in range(len(self.filament_holes[layer_n])):
                self.filament_holes[layer_n][filament_i].add_physical_surface(f"Surface: Filament hole {filament_i} in layer {layer_n}")

        # Matrix: Add physical boundary and surface for each partition
        for i, matrix_partition in enumerate(self.matrix):
            matrix_partition.add_physical_boundary(f"Boundary: Matrix partition {i}")
            matrix_partition.add_physical_surface(f"Surface: Matrix partition {i}")
            
        # Air: 
        self.air.add_physical_boundary(f"Boundary: Air")
        self.air.physical_inner_boundary_tags.append(gmsh.model.addPhysicalGroup(1, [curve.tag for curve in self.air.inner_boundary_curves[0]], name = f"InnerBoundary: Air"))
        self.air.physical_inner_boundary_names.append(f"InnerBoundary: Air")
        self.air.add_physical_surface(f"Surface: Air")

        self.air.strand_bnd_physicalEdgePointTag = gmsh.model.addPhysicalGroup(0, [self.air.inner_boundary_curves[0][0].points[0].tag])

    def save(self, save_file):
        # This function saves the geometry class to a pickle file. 
        # The geometry class is used again during meshing.
        with open(save_file, "wb") as geom_save_file:
            pickle.dump(self, geom_save_file)
            print(f"Geometry saved to file {save_file}")

    def write_geom_to_yaml(self, file_path):
        # This function writes the geometry to a yaml file. 
        # The yaml file contains the coordinates of the points, the type of the curves and the indices of the points that make up the curves and the indices of the curves that make up the areas.
        # Note: Only the strands are written to the yaml file. The air region is not included.
        Conductor = geom.Conductor() # Create a data model for the conductor data

        # Add Materials to the 'Solution' section of the data model
        # 1) Find all unique materials used in the geometry
        # materials = set([surface.material for surface in Surface.surfaces_registry if surface.material is not None])
        # # Sort the materials into two groups by their type. All materials with the same type are grouped together.
        # material_groups = {material_type: [material for material in materials if material.type == material_type] for material_type in set([material.type for material in materials])}
        # # 2) Add all unique materials to the data model, represented by a string with the material type and index in the material group
        # for material_type, material_group in material_groups.items():
        #     for i, material in enumerate(material_group):
        #         material_name = f"{material_type}_{i}"
        #         Conductor.Solution.Materials[material_name] = material
        
        surfaces = list(set(sum(self.filaments, []) + self.matrix)) # Combine all the filaments and the matrix to get all the surfaces which should be written. The air region is not included.
        curves = list(set(sum([surface.boundary_curves for surface in surfaces], []))) # Extract all the boundary curves from the surfaces
        points = list(set(sum([curve.points for curve in curves], []))) # Extract all the points from the curves


        # Populate the points dictionary with coordinates of each point
        # for p, point in enumerate(points):
        #     Conductor.Geometry.Points[p] = geom.Point(Coordinates=point.pos.tolist()) #{"Coordinates": point.pos.tolist()}
        for p, point in enumerate(Point.points_registry):
            if point in points:
                Conductor.Geometry.Points[p] = geom.Point(Coordinates=point.pos.tolist())

        # Populate the curves dictionary with type of each curve and indices of its points
        # for c, curve in enumerate(curves):
        #     curve_points = [Point.points_registry.index(point) for point in curve.points]
        #     Conductor.Geometry.Curves[c] = geom.Curve(
        #         Type=curve.__class__.__name__,
        #         Points=curve_points
        #     )
        for c, curve in enumerate(Curve.curves_registry):
            if curve in curves:
                curve_points = [Point.points_registry.index(point) for point in curve.points]
                Conductor.Geometry.Curves[c] = geom.Curve(
                    Type=curve.__class__.__name__,
                    Points=curve_points
                )


        # Populate the surfaces dictionary with material, boundary curves and inner boundary curves of each surface
        for a, surface in enumerate(Surface.surfaces_registry):
            if surface in surfaces:
                surface_boundary_curves = [Curve.curves_registry.index(curve) for curve in surface.boundary_curves]
                surface_inner_boundary_curves = [[Curve.curves_registry.index(curve) for curve in inner_boundary] for inner_boundary in surface.inner_boundary_curves]

                if surface in self.matrix: # Add dummy values for writing the matrix surfaces to the data model
                    surface.layer = None
                    surface.layer_index = None

                elif surface in sum(self.filaments, []): # Add dummy values for writing the filament surfaces to the data model
                    for l, layer in enumerate(self.filaments):
                        if surface in layer:
                            surface.layer = l
                            surface.layer_index = layer.index(surface)
                            break

                # Name the material based on its type and index in the material groups
                # if surface.material is None:
                #     material_name = None
                # else:
                #     material_type = surface.material.Type
                #     material_index = material_groups[material_type].index(surface.material)
                #     material_name = f"{material_type}_{material_index}"
                material_name = surface.material

                Conductor.Geometry.Areas[a] = geom.Area(
                    Material=material_name,
                    Boundary=surface_boundary_curves,
                    InnerBoundaries=surface_inner_boundary_curves,
                    Layer=surface.layer,
                    LayerIndex=surface.layer_index
                )


        # Write the data model to a yaml file
        UFF.write_data_to_yaml(file_path, Conductor.dict())

    @classmethod
    def read_geom_from_yaml(cls, file_path):
        """
            This function loads a geometry from a yaml file and returns a TwistedStrand object.
            The yaml file contains all points, curves and surfaces which define the geometry.
            - Points are defined by their position vector and can be referenced by an integer ID.
                : Position [x, y, z]
            - Curves are defined by their type (e.g. Line, CircleArc, etc.) and the ID of the points that make up the curves. Curves are referenced by an integer ID as well.
                : Type ('Line', 'CircleArc', etc.)
                : Points ([1,2,3]), list of point-ID defining the curve.
            - Surfaces are defined by material, outer boundary, inner boundaries and layer and layer-index if the surface is a strand. 
                : Material ('Cu', 'NbTi', 'Nb3Sn', etc.)
                : Outer boundary ([2,3,4,5...]), list of curve-ID defining the outer boundary closed loop.
                : Inner boundaries ([[1,2,3], [4,5,6], ... ]), list of list of curve-IDs defining closed loops.
                : Layer (0, 1, 2, ...), the layer of a filament. None if the surface is part of the matrix.
                : LayerIndex (0, 1, 2, ...), the index of the filament in the layer. None if the surface is part of the matrix.

            :param file_path: The full path to the yaml file.
            :type file_path: str
            :param gmsh_curve_convention: If True, the curves are created using the gmsh convention for defining curves. Determines the order of the points in the curves. Defaults to False. This is a temporary solution. In the future, the order of the points in the curves will be updated to fit the gmsh convention.

        """
        Conductor = UFF.read_data_from_yaml(file_path, geom.Conductor)
        
        # 1) Create the points
        for point in Conductor.Geometry.Points.values():
            Point.create_or_get(point.Coordinates)

        # 2) Create the curves
        for curve in Conductor.Geometry.Curves.values():
            curve_type = curve.Type
            points = [Point.points_registry[p] for p in curve.Points]


            c = globals()[curve_type].create_or_get(*points) # Create the curve of the specified type

            # TODO: To be added.
            # c.contact = curve.Contact 
            # c.thickness = curve.Thickness
            # c.material = curve.Material

        
        # 3) Create the surfaces
        # TODO: area.boundary_material and boundary_thickness are not yet used.
        strand = cls()
        layers = max([area.Layer for area in Conductor.Geometry.Areas.values() if area.Layer is not None]) + 1 # The number of layers in the strand
        strand.filaments = [[None for i in range(6)] for j in range(layers)] # Initialize the filaments list
        strand.filament_holes = [[None for i in range(6)] for j in range(layers)] # Initialize the filament holes list
        
        for area_index, area_dm in Conductor.Geometry.Areas.items():
            boundary_curves = [Curve.curves_registry[c] for c in area_dm.Boundary]
            inner_boundary_curves = [[Curve.curves_registry[c] for c in inner_boundary] for inner_boundary in area_dm.InnerBoundaries]
            surface = Surface(boundary_curves, inner_boundary_curves)
            if area_dm.Material:
                surface.material = area_dm.Material



            if area_dm.Layer is None:
                # It is either a matrix partition or it is a hole in a filament. 
                # We check if it is a hole in a filament by checking if the area outer boundary is in the inner boundary of a filament.
                is_hole = False
                for other_area in Conductor.Geometry.Areas.values(): # Loop over all areas to check if the area is a hole in a filament
                    if other_area.Layer is not None: # If the other area is a filament
                        if area_dm.Boundary in other_area.InnerBoundaries: # If the area is a hole in the other filament
                            # boundary_curves[0].P1, boundary_curves[0].P2 = boundary_curves[0].P2, boundary_curves[0].P1 # Reverse the order of the boundary curve points to get the correct orientation
                            layer = other_area.Layer
                            layer_index = other_area.LayerIndex
                            strand.filament_holes[layer][layer_index] = surface
                            is_hole = True
                            break
                
                if not is_hole: # If it is not a hole, it is a matrix partition
                    strand.matrix.append(surface)

            else:
                strand.filaments[area_dm.Layer][area_dm.LayerIndex] = surface
        
        # Remove None values from the filaments list
        strand.filaments = [[filament for filament in layer if filament is not None] for layer in strand.filaments]
        strand.filament_holes = [[hole for hole in layer if hole is not None] for layer in strand.filament_holes]


        # Sort the matrix partitions from inner to outer based on the outermost point of the boundary curves of the partitions
        strand.matrix = sorted(strand.matrix, key=lambda surface: max([max([np.linalg.norm(point.pos) for point in curve.points]) for curve in surface.boundary_curves]))
        
        return strand



        

class Geometry:
    """
    Class to generate the ConductorAC Strand geometry.

    This class is responsible for generating the geometry of the twisted strand. 
    It can either load a geometry from a YAML file or create the model from scratch. 
    The geometry is saved to a .brep file and the geometry class is saved as a pickle file. If specified, the geometry representation can also be saved to a YAML file.

    :ivar fdm: The fiqus inputs data model.
    :vartype fdm: object
    :ivar cacdm: The magnet section of the fiqus inputs data model.
    :vartype cacdm: object
    :ivar inputs_folder: The full path to the folder with input files, i.e., conductor and STEP files.
    :vartype inputs_folder: str
    :ivar geom_folder: The full path to the current working directory.
    :vartype geom_folder: str
    :ivar magnet_name: The name of the magnet.
    :vartype magnet_name: str
    :ivar geom_file: The full path to the .brep file where the geometry will be saved.
    :vartype geom_file: str
    :ivar verbose: If True, more information is printed in the Python console.
    :vartype verbose: bool
    :ivar gu: An instance of the GmshUtils class.
    :vartype gu: object
    """

    def __init__(self, fdm, inputs_folder_path, verbose=True):
        """
        Initializes the Geometry class.

        :param fdm: The fiqus data model.
        :type fdm: object
        :param inputs_folder_path: The full path to the folder with input files, i.e., conductor and STEP files.
        :type inputs_folder_path: str
        :param verbose: If True, more information is printed in the Python console. Defaults to True.
        :type verbose: bool, optional
        """
        self.fdm = fdm
        self.cacdm = fdm.magnet
        self.inputs_folder = inputs_folder_path
        self.geom_folder = os.path.join(os.getcwd())
        self.magnet_name = fdm.general.magnet_name
        self.geom_file = os.path.join(self.geom_folder, f'{self.magnet_name}.brep')
        self.verbose = verbose
        self.gu = GmshUtils(self.geom_folder, self.verbose)
        self.gu.initialize()

        # To see the surfaces in a better way in GUI:
        gmsh.option.setNumber("Geometry.SurfaceType", 2)

    def generate_strand_geometry(self, gui=False):
        """
        Generates the geometry of a strand based on the settings specified in the input data model.
        The geometry can either be loaded from a YAML file or created from scratch. After creation, the geometry
        can be saved to a YAML file. The method also creates gmsh instances, adds physical groups to the geometry,
        writes the geometry to a .brep file, and saves the geometry-class to a pickle file. If `gui` is True, it
        launches an interactive GUI.

        :param gui: If True, launches an interactive GUI after generating the geometry. Default is False.
        :type gui: bool, optional

        :return: None
        """
        print("Generating geometry")
        #0) Clear the registries. Used when generating reference files for tests.
        if Point.points_registry: # If the points registry is not empty, clear it.
            Point.points_registry.clear()
        if Curve.curves_registry: # If the curves registry is not empty, clear it.
            Curve.curves_registry.clear()
        if Surface.surfaces_registry: # If the surfaces registry is not empty, clear it.
            Surface.surfaces_registry.clear()
        if Surface.curve_loop_registry: # If the curve loop registry is not empty, clear it.
            Surface.curve_loop_registry.clear()

        # 1) Either load the geometry from a yaml file or create the model from scratch
        if self.cacdm.geometry.io_settings.load.load_from_yaml:
            CAC = TwistedStrand.read_geom_from_yaml(os.path.join(self.inputs_folder, self.cacdm.geometry.io_settings.load.filename))
        else:
            CAC = TwistedStrand()
            conductor_name = self.cacdm.solve.conductor_name
            CAC.create_geometry(
                self.fdm.conductors[conductor_name].strand.filament_diameter/2,
                self.cacdm.geometry.hexagonal_filaments,
                self.fdm.conductors[conductor_name].strand.number_of_filaments,
                self.fdm.conductors[conductor_name].strand.diameter_core/2, 
                self.fdm.conductors[conductor_name].strand.diameter_filamentary/2,
                self.fdm.conductors[conductor_name].strand.diameter/2,
                self.cacdm.geometry.filament_circular_distribution
            )
        CAC.add_air(self.cacdm.geometry.air_radius)
        CAC.create_gmsh_instance()
        
        # 2) Save the geometry to a yaml file if specified
        if self.cacdm.geometry.io_settings.save.save_to_yaml: 
            filename = self.cacdm.geometry.io_settings.save.filename
            CAC.write_geom_to_yaml(os.path.join(self.geom_folder, filename))
        

        gmsh.model.occ.synchronize()

        CAC.add_physical_groups() # Add physical groups to the geometry
        print("Writing geometry")
        gmsh.write(self.geom_file) # Write the geometry to a .brep file

        CAC.save(os.path.join(self.geom_folder, f'{self.magnet_name}.pkl')) # Save the geometry-class to a pickle file

        if gui:
            self.gu.launch_interactive_GUI()
 
    def load_conductor_geometry(self, gui=False):
        """
        Loads geometry from .brep file.
        """

        print("Loading geometry")

        gmsh.clear()
        gmsh.model.occ.importShapes(self.geom_file, format="brep")
        gmsh.model.occ.synchronize()

        if gui:
            self.gu.launch_interactive_GUI()








from pydantic import BaseModel, Field
from typing import Dict, List, Union, Literal, Optional
from fiqus.data.DataRoxieParser import RoxieData
from fiqus.data.DataFiQuSCCT import CCTDM
from fiqus.data.DataFiQuSMultipole import Multipole
from fiqus.data.DataFiQuSPancake3D import Pancake3D
from fiqus.data.DataConductor import Conductor
from fiqus.data.DataFiQuSConductorAC_Strand import CACStrand


class FiQuSGeometry(BaseModel):
    """
    Class for Roxie data
    """

    Roxie_Data: RoxieData = RoxieData()


class RunFiQuS(BaseModel):
    """
    Class for FiQuS run
    """

    type: Literal[
        "start_from_yaml",
        "mesh_only",
        "geometry_only",
        "geometry_and_mesh",
        "pre_process_only",
        "mesh_and_solve_with_post_process_python",
        "solve_with_post_process_python",
        "solve_only",
        "post_process_getdp_only",
        "post_process_python_only",
        "post_process",
        "plot_python",
        "batch_post_process_python",
    ] = Field(
        default="start_from_yaml",
        title="Run Type of FiQuS",
        description="FiQuS allows you to run the model in different ways. The run type can be specified here. For example, you can just create the geometry and mesh or just solve the model with previous mesh, etc.",
    )
    geometry: Optional[Union[str, int]] = Field(
        default=None,
        title="Geometry Folder Key",
        description="This key will be appended to the geometry folder.",
    )
    mesh: Optional[Union[str, int]] = Field(
        default=None,
        title="Mesh Folder Key",
        description="This key will be appended to the mesh folder.",
    )
    solution: Optional[Union[str, int]] = Field(
        default=None,
        title="Solution Folder Key",
        description="This key will be appended to the solution folder.",
    )
    launch_gui: bool = Field(
        default=False,
        title="Launch GUI",
        description="If True, the GUI will be launched after the run.",
    )
    overwrite: bool = Field(
        default=False,
        title="Overwrite",
        description="If True, the existing folders will be overwritten, otherwise new folders will be created. NOTE: This setting has no effect for HTCondor runs.",
    )
    comments: str = Field(
        default="",
        title="Comments",
        description="Comments for the run. These comments will be saved in the run_log.csv file.",
    )
    verbosity_Gmsh: int = Field(
        default=5,
        title="verbosity_Gmsh",
        description="Level of information printed on the terminal and the message console (0: silent except for fatal errors, 1: +errors, 2: +warnings, 3: +direct, 4: +information, 5: +status, 99: +debug)",
    )
    verbosity_GetDP: int = Field(
        default=5,
        title="verbosity_GetDP",
        description="Level of information printed on the terminal and the message console. Higher number prints more, good options are 5 or 6.",
    )
    verbosity_FiQuS: bool = Field(
        default=True,
        title="verbosity_FiQuS",
        description="Level of information printed on the terminal and the message console by FiQuS. Only True of False for now.",
    )


class GeneralFiQuS(BaseModel):
    """
    Class for FiQuS general
    """

    magnet_name: Optional[str] = None


class EnergyExtraction(BaseModel):
    """
    Level 3: Class for FiQuS
    """

    t_trigger: Optional[float] = None
    R_EE: Optional[float] = None
    power_R_EE: Optional[float] = None
    L: Optional[float] = None
    C: Optional[float] = None


class QuenchHeaters(BaseModel):
    """
    Level 3: Class for FiQuS
    """

    N_strips: Optional[int] = None  # set to 0 to avoid building quench heater thin shells
    t_trigger: Optional[List[float]] = None
    U0: Optional[List[float]] = None
    C: Optional[List[float]] = None
    R_warm: Optional[List[float]] = None
    w: Optional[List[float]] = None
    h: Optional[List[float]] = None
    h_ins: List[List[float]] = []
    type_ins: List[List[str]] = []
    h_ground_ins: List[List[float]] = []
    type_ground_ins: List[List[str]] = []
    l: Optional[List[float]] = None
    l_copper: Optional[List[float]] = None
    l_stainless_steel: Optional[List[float]] = None
    ids: Optional[List[int]] = None
    turns: Optional[List[int]] = None
    turns_sides: Optional[List[str]] = None


class Cliq(BaseModel):
    """
    Level 3: Class for FiQuS
    """

    t_trigger: Optional[float] = None
    current_direction: Optional[List[int]] = None
    sym_factor: Optional[int] = None
    N_units: Optional[int] = None
    U0: Optional[float] = None
    C: Optional[float] = None
    R: Optional[float] = None
    L: Optional[float] = None
    I0: Optional[float] = None


class Circuit(BaseModel):
    """
    Level 2: Class for FiQuS
    """

    R_circuit: Optional[float] = None
    L_circuit: Optional[float] = None
    R_parallel: Optional[float] = None


class PowerSupply(BaseModel):
    """
    Level 2: Class for FiQuS
    """

    I_initial: Optional[float] = None
    t_off: Optional[float] = None
    t_control_LUT: Optional[List[float]] = Field(
        default=None,
        title="Time Values for Current Source",
        description="This list of time values will be matched with the current values in I_control_LUT, and then these (t, I) points will be connected with straight lines.",
    )
    I_control_LUT: Optional[List[float]] = Field(
        default=None,
        title="Current Values for Current Source",
        description="This list of current values will be matched with the time values in t_control_LUT, and then these (t, I) points will be connected with straight lines.",
    )
    R_crowbar: Optional[float] = None
    Ud_crowbar: Optional[float] = None


class QuenchProtection(BaseModel):
    """
    Level 2: Class for FiQuS
    """

    energy_extraction: EnergyExtraction = EnergyExtraction()
    quench_heaters: QuenchHeaters = QuenchHeaters()
    cliq: Cliq = Cliq()

class QuenchDetection(BaseModel):
    """
    Level 2: Class for FiQuS
    """

    voltage_thresholds: Optional[List[float]] = Field(
        default=None,
        title="List of quench detection voltage thresholds",
        description="Voltage thresholds for quench detection. The quench detection will be triggered when the voltage exceeds these thresholds continuously for a time larger than the discrimination time.",
    )

    discrimination_times: Optional[List[float]] = Field(
        default=None,
        title="List of quench detection discrimination times",
        description="Discrimination times for quench detection. The quench detection will be triggered when the voltage exceeds the thresholds continuously for a time larger than these discrimination times.",
    )

    voltage_tap_pairs: Optional[List[List[int]]] = Field(
        default=None,
        title="List of quench detection voltage tap pairs",
        description="Voltage tap pairs for quench detection. The voltage difference between these pairs will be used for quench detection.",
    )

class FDM(BaseModel):
    """
    Class for FiQuS
    """

    general: GeneralFiQuS = GeneralFiQuS()
    run: RunFiQuS = RunFiQuS()
    magnet: Union[Multipole, CCTDM, Pancake3D, CACStrand] = Field(
        default=Multipole(), discriminator="type"
    )
    circuit: Circuit = Circuit()
    power_supply: PowerSupply = PowerSupply()
    quench_protection: QuenchProtection = QuenchProtection()
    quench_detection: QuenchDetection = QuenchDetection()
    conductors: Dict[Optional[str], Conductor] = {}



from pydantic import BaseModel
from typing import Dict, List, Tuple, Optional


class Coord(BaseModel):
    x: Optional[float] = None
    y: Optional[float] = None


class Area(BaseModel):
    loop: Optional[int] = None
    surface: Optional[int] = None


class Region(BaseModel):
    points: Dict[str, int] = {}
    lines: Dict[str, int] = {}
    areas: Dict[str, Area] = {}


# class CableInsulated(BaseModel):
#     type: #Literal['Insulated']
#     insulated: Region = Region()
#
#
# class CableBare(BaseModel):
#     type: #Literal['Bare']
#     bare: Region = Region()
#     insulation: Region = Region()


class BlockData(BaseModel):
    half_turns: Region = (
        Region()
    )  # Union[CableInsulated, CableBare] = {'type': 'Insulated'}
    insulation: Region = Region()
    current_sign: Optional[int] = None


class Block(BaseModel):
    blocks: Dict[int, BlockData] = {}
    conductor_name: Optional[str] = None
    conductors_number: Optional[int] = None


class Winding(BaseModel):
    windings: Dict[int, Block] = {}


class Layer(BaseModel):
    layers: Dict[int, Winding] = {}


class Pole(BaseModel):
    poles: Dict[int, Layer] = {}
    type: Optional[str] = None
    bore_center: Coord = Coord()


class Order(BaseModel):
    coil: Optional[int] = None
    pole: Optional[int] = None
    layer: Optional[int] = None
    winding: Optional[int] = None
    block: Optional[int] = None


class AnticlockwiseOrder(BaseModel):
    pole: Optional[int] = None
    winding: Optional[int] = None
    block: Optional[int] = None


class MidLayer(BaseModel):
    half_turn_lists: Dict[str, List[int]] = {}
    point_angles: Dict[str, float] = {}
    mid_layers: Region = Region()


class Iron(BaseModel):
    quadrants: Dict[int, Region] = {}
    max_radius: float = 0.0


class LayerOrder(BaseModel):
    layers: Dict[int, List[AnticlockwiseOrder]] = {}


class CoilOrder(BaseModel):
    coils: Dict[int, LayerOrder] = {}


class Coil(BaseModel):
    coils: Dict[int, Pole] = {}
    physical_order: List[Order] = []
    anticlockwise_order: CoilOrder = CoilOrder()
    concentric_coils: Dict[Tuple[float, float], List[int]] = {}
    max_radius: float = 0.0


class InsulationThickness(BaseModel):
    mid_pole: Dict[str, float] = {}
    mid_layer: Dict[str, float] = {}
    mid_winding: Dict[str, float] = {}


class ThinShell(BaseModel):
    mid_layers_ht_to_ht: Dict[str, MidLayer] = {}
    mid_layers_wdg_to_ht: Dict[str, Region] = {}
    mid_layers_ht_to_wdg: Dict[str, Region] = {}
    mid_layers_wdg_to_wdg: Dict[str, Region] = {}
    mid_poles: Dict[str, Region] = {}
    mid_windings: Dict[str, Region] = {}
    mid_turn_blocks: Dict[str, Region] = {}
    mid_wedge_turn: Dict[str, Region] = {}
    mid_layers_aux: Dict[str, Region] = {}
    ins_thickness: InsulationThickness = InsulationThickness()


class WedgeRegion(BaseModel):
    wedges: Dict[int, Region] = {}
    block_prev: Dict[int, int] = {}
    block_next: Dict[int, int] = {}


class WedgeLayer(BaseModel):
    layers: Dict[int, WedgeRegion] = {}


class Wedge(BaseModel):
    coils: Dict[int, WedgeLayer] = {}


class InsulationRegion(BaseModel):
    ins: Region = Region()
    blocks: List[List[int]] = []
    wedges: List[List[int]] = []


class InsulationGroup(BaseModel):
    group: Dict[int, InsulationRegion] = {}


class Insulation(BaseModel):
    coils: Dict[int, InsulationGroup] = {}


class Geometry(BaseModel):
    coil: Coil = Coil()
    iron: Iron = Iron()
    wedges: Wedge = Wedge()
    air: Region = Region()
    air_inf: Region = Region()
    symmetric_boundaries: Region = Region()
    thin_shells: ThinShell = ThinShell()
    insulation: Insulation = Insulation()


# Domain classes #
class MidLayerLine(BaseModel):
    inner: Dict[str, int] = {}
    outer: Dict[str, int] = {}


class GroupType(BaseModel):
    curves: Dict[str, int] = {}
    surfaces: Dict[str, int] = {}


class PoweredGroup(BaseModel):
    tag: Optional[int] = None
    group: Optional[str] = None
    lines: Dict[str, int] = {}
    mid_layer_lines: MidLayerLine = MidLayerLine()
    mid_pole_lines: Dict[str, int] = {}
    mid_winding_lines: Dict[str, int] = {}
    mid_turn_lines: Dict[str, int] = {}
    aux_lines: Dict[str, int] = {}


class WedgeGroup(BaseModel):
    tag: Optional[int] = None
    group: Optional[str] = None
    lines: Dict[str, int] = {}
    mid_layer_lines: MidLayerLine = MidLayerLine()
    mid_turn_lines: Dict[str, int] = {}
    aux_lines: Dict[str, int] = {}


class PoweredBlock(BaseModel):
    half_turns: Dict[int, PoweredGroup] = {}
    current_sign: Optional[int] = None
    conductor: Optional[str] = None


class SymmetryGroup(BaseModel):
    normal_free: Optional[int] = None
    tangential_free: Optional[int] = None


class PhysicalGroup(BaseModel):
    blocks: Dict[int, PoweredBlock] = {}
    wedges: Dict[int, WedgeGroup] = {}
    insulations: GroupType = GroupType()
    iron: GroupType = GroupType()
    air_inf: Optional[int] = None
    air_inf_bnd: Optional[int] = None
    air: Optional[int] = None
    symmetric_boundaries: SymmetryGroup = SymmetryGroup()
    half_turns_points: Optional[int] = None


class SymmetryBoundaries(BaseModel):
    x: List[int] = []
    y: List[int] = []


class GroupEntities(BaseModel):
    iron: Dict[str, List[int]] = {}
    air: List[int] = []
    symmetric_boundaries: SymmetryBoundaries = SymmetryBoundaries()


class Domain(BaseModel):
    groups_entities: GroupEntities = GroupEntities()
    physical_groups: PhysicalGroup = PhysicalGroup()


class MultipoleData(BaseModel):
    geometries: Geometry = Geometry()
    domains: Domain = Domain()


#######################################


class NeighbourNode(BaseModel):
    groups: Dict[str, List[float]] = {}


class IsothermalNodes(BaseModel):
    conductors: Dict[str, Dict[int, List[float]]] = {}
    wedges: Dict[str, Dict[int, List[float]]] = {}
    thin_shells: Dict[int, List[float]] = {}


class MultipoleRegionCoordinate(BaseModel):
    isothermal_nodes: IsothermalNodes = IsothermalNodes()
    neighbouring_nodes: NeighbourNode = NeighbourNode()


# if __name__ == "__main__":
#     write = True
#     read = False
#
#     def read_regions(regions_file_name):
#         with open(regions_file_name, 'r') as stream:
#             yaml_str = ruamel.yaml.safe_load(stream)
#         return MultipoleData(**yaml_str)
#
#     if write:
#         model = MultipoleData()
#         with open('FiQuS_data.yaml', 'w') as yaml_file:
#             ruamel.yaml.dump(model.dict(), yaml_file, default_flow_style=False)



import math
from enum import Enum
from typing import List, Tuple, Dict

import os
import numpy as np
import gmsh
import copy

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

from fiqus.parsers.ParserCOND import ParserCOND

class direction(Enum):
    """
    A class to specify direction easily.
    """

    ccw = 0
    cw = 1

class coordinate(Enum):
    """
    A class to specify coordinate types easily.
    """

    rectangular = 0
    cylindrical = 1
    spherical = 2

class point:
    """
    This is a class for creating points in GMSH. It supports rectangular and cylindrical
    coordinates. Moreover, vector operations are supported.

    :param r0: x, r, or r (default: 0.0)
    :type r0: float, optional
    :param r1: y, theta, or theta (default: 0.0)
    :type r1: float, optional
    :param r2: z, z, or phi (default: 0.0)
    :type r2: float, optional
    :param type: coordinate type (default: coordinate.rectangular)
    :type type: coordinate, optional
    """

    def __init__(self, r0=0.0, r1=0.0, r2=0.0, type=coordinate.rectangular) -> None:

        self.type = type  # Store 'type' as an instance attribute

        if type is coordinate.rectangular:
            self.x = r0
            self.y = r1
            self.z = r2

            self.r = math.sqrt(self.x**2 + self.y**2)
            self.theta = math.atan2(self.y, self.x)
        elif type is coordinate.cylindrical:
            self.r = r0
            self.theta = r1
            self.x = self.r * math.cos(self.theta)
            self.y = self.r * math.sin(self.theta)
            self.z = r2
        elif type is coordinate.spherical:
            raise ValueError("Spherical coordinates are not supported yet!")
        else:
            raise ValueError("Improper coordinate type value!")

        self.tag = gmsh.model.occ.addPoint(self.x, self.y, self.z)

    def __repr__(self):
        """
        Returns the string representation of the point.

        :return: string representation of the point
        :rtype: str
        """
        return "point(%r, %r, %r, %r)" % (self.x, self.y, self.z, self.type)

    def __abs__(self):
        """
        Returns the magnitude of the point vector.

        :return: the magnitude of the point vector
        :rtype: float
        """
        return math.hypot(self.x, self.y, self.z)

    def __add__(self, other):
        """
        Returns the summation of two point vectors.

        :param other: point vector to be added
        :type other: point
        :return: the summation of two point vectors
        :rtype: point
        """
        x = self.x + other.x
        y = self.y + other.y
        z = self.z + other.z
        return point(x, y, z, coordinate.rectangular)

    def __mul__(self, scalar):
        """
        Returns the product of a point vector and a scalar.

        :param scalar: a scalar value
        :type scalar: float
        :return: point
        :rtype: point
        """
        return point(
            self.x * scalar,
            self.y * scalar,
            self.z * scalar,
            coordinate.rectangular,
        )



class spiralCurve:
    """
    A class to create a spiral curves parallel to XY plane in GMSH. The curve is defined
    by a spline and it is divided into sub-curves. Sub-curves are used because it makes
    the geometry creation process easier.

    :param innerRadius: inner radius
    :type innerRadius: float
    :param gap: gap after each turn
    :type gap: float
    :param turns: number of turns
    :type turns: float
    :param z: z coordinate
    :type z: float
    :param initialTheta: initial theta angle in radians
    :type initialTheta: float
    :param direction: direction of the spiral (default: direction.ccw)
    :type direction: direction, optional
    :param cutPlaneNormal: normal vector of the plane that will cut the spiral curve
        (default: None)
    :type cutPlaneNormal: tuple[float, float, float], optional
    """

    # If the number of points used per turn is n, then the number of sections per turn
    # is n-1. They set the resolution of the spiral curve. It sets the limit of the
    # precision of the float number of turns that can be used to create the spiral
    # curve. The value below might be modified in Geometry.__init__ method.
    sectionsPerTurn = 16

    # There will be curvesPerTurn curve entities per turn. It will be effectively the
    # number of volumes per turn in the end. The value below might be modified in
    # Geometry.__init__ method.
    curvesPerTurn = 2

    def __init__(
        self,
        innerRadius,
        gap,
        turns,
        z,
        initialTheta,
        transitionNotchAngle,
        geo,
        direction=direction.ccw, # TODO code this to understand cw direction
        cutPlaneNormal=Tuple[float, float, float]
    ) -> None:
        spt = self.sectionsPerTurn  # just to make the code shorter
        self.turnRes = 1 / spt  # turn resolution
        cpt = self.curvesPerTurn  # just to make the code shorter
        self.turns = turns
        self.geo = geo
        # =============================================================================
        # GENERATING POINTS STARTS ====================================================
        # =============================================================================
        print('The theta is')
        print(initialTheta)

        print('This is the status of the cutPlane')
        print(cutPlaneNormal)
        # Calculate the coordinates of the points that define the spiral curve:
        if direction is direction.ccw:
            # If the spiral is counter-clockwise, the initial theta angle decreases,
            # and r increases as the theta angle decreases.
            multiplier = 1
        elif direction is direction.cw:
            # If the spiral is clockwise, the initial theta angle increases, and r
            # increases as the theta angle increases.
            multiplier = -1

        NofPointsPerTurn = int(spt + 1)
        thetaArrays = []
        for turn in range(1, int(self.turns) + 1):
            thetaArrays.append(
                np.linspace(
                    initialTheta + (turn - 1) * 2 * math.pi * multiplier,
                    initialTheta + (turn) * 2 * math.pi * multiplier,
                    NofPointsPerTurn,
                )
            )

        thetaArrays.append(
            np.linspace(
                initialTheta + (turn) * 2 * math.pi * multiplier,
                initialTheta + (self.turns) * 2 * math.pi * multiplier,
                round(spt * (self.turns - turn) + 1),
            )
        )

        if cutPlaneNormal is not None:
            # If the cutPlaneNormal is specified, the spiral curve will be cut by a
            # plane that is normal to the cutPlaneNormal vector and passes through the
            # origin.

            alpha = math.atan2(cutPlaneNormal[1], cutPlaneNormal[0]) - math.pi / 2
            alpha2 = alpha + math.pi

            listOfBreakPoints = []
            for turn in range(1, int(self.turns) + 2):
                breakPoint1 = alpha + (turn - 1) * 2 * math.pi * multiplier
                breakPoint2 = alpha2 + (turn - 1) * 2 * math.pi * multiplier
                if (
                    breakPoint1 > initialTheta
                    and breakPoint1 < initialTheta + 2 * math.pi * self.turns
                ):
                    listOfBreakPoints.append(breakPoint1)
                if (
                    breakPoint2 > initialTheta
                    and breakPoint2 < initialTheta + 2 * math.pi * self.turns
                ):
                    listOfBreakPoints.append(breakPoint2)

            thetaArrays.append(np.array(listOfBreakPoints))

        theta = np.concatenate(thetaArrays)
        theta = np.round(theta, 10)
        theta = np.unique(theta)
        theta = np.sort(theta)
        theta = theta[::multiplier]

        r = innerRadius + (theta - initialTheta) / (2 * math.pi) * (gap) * multiplier
        z = np.ones(theta.shape) * z

        # Create the points and store their tags:
        points = []  # point objects
        pointTags = []  # point tags
        breakPointObjectsDueToCutPlane = []  # only used if cutPlaneNormal is not None
        breakPointTagsDueToCutPlane = []  # only used if cutPlaneNormal is not None
        pointObjectsWithoutBreakPoints = []  # only used if cutPlaneNormal is not None
        pointTagsWithoutBreakPoints = []  # only used if cutPlaneNormal is not None
        breakPointObjectsDueToTransition = []
        breakPointTagsDueToTransition = []
        coordinateList = []

        for j in range(len(theta)):
            pointObject = point(r[j], theta[j], z[j], coordinate.cylindrical)
            [x_c, y_c, z_c] = [r[j], theta[j], z[j]]
            coordinateList.append([x_c, y_c, z_c])
            points.append(pointObject)
            pointTags.append(pointObject.tag)
            if cutPlaneNormal is not None:
                if theta[j] in listOfBreakPoints:
                    breakPointObjectsDueToCutPlane.append(pointObject)
                    breakPointTagsDueToCutPlane.append(pointObject.tag)
                else:
                    pointObjectsWithoutBreakPoints.append(pointObject)
                    pointTagsWithoutBreakPoints.append(pointObject.tag)

            # identify if the point is a break point due to the layer transition:
            angle1 = initialTheta + (2 * math.pi - transitionNotchAngle) * multiplier
            angle2 = (
                initialTheta
                + ((self.turns % 1) * 2 * math.pi + transitionNotchAngle) * multiplier
            )
            if math.isclose(
                math.fmod(theta[j], 2 * math.pi), angle1, abs_tol=1e-6
            ) or math.isclose(math.fmod(theta[j], 2 * math.pi), angle2, abs_tol=1e-6):
                breakPointObjectsDueToTransition.append(pointObject)
                breakPointTagsDueToTransition.append(pointObject.tag)

        # Logic to break the points up into relevant geom coordinates
        # Brick points structure (for X-Y plane only for now):
        # [[[x1, y1, z1], [x2, y2, z2], [x3, y3, z3], [x4, y4, z4]], ...]
        # Theoretically, very easy to extend to 8 points knowing the height of the


        # Defining the coordinate lists to which the points are to be added

        # winding one covers the list of points in the domain of theta [k, pi*k], where k is an integer number
        winding_1 = []
        # winding one covers the list of points in the domain of theta [pi*k, 2pi*k], where k is an integer number
        winding_2 = []
        # winding one covers the list of points in the domain of theta [k, pi*k], where k is an integer number
        winding_3 = []
        # winding one covers the list of points in the domain of theta [pi*k, 2pi*k], where k is an integer number
        winding_4 = []

        for i in range(len(theta)-1): # range is reduced as no brick can be created starting at the last point
            # print(i)
            # Assuming theta is a numpy array and you're looking for the index of a value close to pi
            value_to_find = theta[i]+2*np.pi
            tolerance = 1e-10  # Define a small tolerance
            # Find indices where the condition is true
            indices = np.where(np.abs(theta - value_to_find) < tolerance)[0]
            z = self.geo.wi.h/2
            z_p = z
            z_n = -z

            if len(indices) > 0:
                windingUpIndex = indices[0]  # Take the first index if there are multiple matches

                try:
                    x_1 = r[i] * np.cos(theta[i])
                    y_1 = r[i] * np.sin(theta[i])
                    x_2 = r[i + 1] * np.cos(theta[i + 1])
                    y_2 = r[i + 1] * np.sin(theta[i + 1])
                    x_3 = r[windingUpIndex] * np.cos(theta[windingUpIndex])
                    y_3 = r[windingUpIndex] * np.sin(theta[windingUpIndex])
                    x_4 = r[windingUpIndex + 1] * np.cos(theta[windingUpIndex + 1])
                    y_4 = r[windingUpIndex + 1] * np.sin(theta[windingUpIndex + 1])

                    addPoints = [[x_1, y_1, z_n], [x_2, y_2, z_n], [x_4, y_4, z_n], [x_3, y_3, z_n], [x_1, y_1, z_p], [x_2, y_2, z_p], [x_4, y_4, z_p], [x_3, y_3, z_p]]

                    k = int(theta[i] / (2 * np.pi))

                    angle_in_current_turn = round(theta[i] % (2 * np.pi), 6)
                    angle_second_brick_point = round(theta[i+1] % (2 * np.pi), 6)

                    if (((round(angle_in_current_turn, 6) <= round(np.pi, 6)) or (round(angle_in_current_turn, 6) == round(2*np.pi, 6))) and (round(angle_second_brick_point, 6) <= round(np.pi, 6))):
                        k = int(round(theta[i] / (2 * np.pi)))
                        if k % 2 == 0:
                            winding_1.append(addPoints)
                        else:
                            winding_3.append(addPoints)

                    if (round(angle_in_current_turn, 6) >= round(np.pi, 6) and (round(angle_second_brick_point, 6) >= round(np.pi, 6) or round(angle_second_brick_point, 6) == 0.0)):

                        if k % 2 == 0:
                            winding_2.append(addPoints)
                        else:
                            winding_4.append(addPoints)

                except IndexError:
                    print('All of the winding conductor points have been found')

        x_coords = []
        y_coords = []
        z_coords = []

        # Writing the conductor file
        windingPointList = [winding_1, winding_2, winding_3, winding_4]

        if self.geo.conductorWrite:
            dict_cond_sample = {0: {'SHAPE': 'BR8', 'XCENTRE': '0.0', 'YCENTRE': '0.0', 'ZCENTRE': '0.0', 'PHI1': '0.0', 'THETA1': '0.0', 'PSI1': '0.0', 'XCEN2': '0.0', 'YCEN2': '0.0', 'ZCEN2': '0.0', 'THETA2': '0.0', 'PHI2': '0.0', 'PSI2': '0.0', 'XP1': '-0.879570', 'YP1': '-0.002940', 'ZP1': '-1.131209', 'XP2': '-0.879570', 'YP2': '0.002940', 'ZP2': '-1.131209', 'XP3': '-0.881381', 'YP3': '0.002940', 'ZP3': '-1.114205', 'XP4': '-0.881381', 'YP4': '-0.002940', 'ZP4': '-1.114205', 'XP5': '-0.861227', 'YP5': '-0.002972', 'ZP5': '-1.129183', 'XP6': '-0.861208', 'YP6': '0.002908', 'ZP6': '-1.129182', 'XP7': '-0.863294', 'YP7': '0.002912', 'ZP7': '-1.112210', 'XP8': '-0.863313', 'YP8': '-0.002968', 'ZP8': '-1.112211', 'CURD': '201264967.975494', 'SYMMETRY': '1', 'DRIVELABEL': 'drive 0', 'IRXY': '0', 'IRYZ': '0', 'IRZX': '0', 'TOLERANCE': '1e-6'}}

            # Use a dictionary to manage dict_cond_1, dict_cond_2, etc.
            dict_conds = {}

            k = 1  # Start from 1 for dict_cond_1, dict_cond_2, ...
            for winding in windingPointList:
                n = 0
                dict_conds[k] = {}
                for brick in winding:
                    dict_conds[k][n] = copy.deepcopy(dict_cond_sample[0])
                    for pointIndex in range(8):
                        dict_conds[k][n][f'XP{pointIndex + 1}'] = str(brick[pointIndex][0])
                        dict_conds[k][n][f'YP{pointIndex + 1}'] = str(brick[pointIndex][1])
                        dict_conds[k][n][f'ZP{pointIndex + 1}'] = str(brick[pointIndex][2])
                    n += 1
                k += 1

            target_dir = os.path.join(os.getcwd(), 'tests', '_outputs', 'parsers')

            # Ensure the target directory exists
            os.makedirs(target_dir, exist_ok=True)
            for i in range (1, 5):
                # Define the output file path
                out_file = os.path.join(target_dir, f'winding_{i}.cond')
                input_dict = dict_conds[i]
                list_of_shapes = ['BR8']
                for shape in list_of_shapes:
                        pc = ParserCOND()
                        print(f'the input dictionary is {input_dict}')
                        pc.write_cond(input_dict, out_file)

        # =============================================================================
        # GENERATING POINTS ENDS ======================================================
        # =============================================================================

        # =============================================================================
        # GENERATING SPLINES STARTS ===================================================
        # =============================================================================

        # Create the spline with the points:
        spline = gmsh.model.occ.addSpline(pointTags)

        # Split the spline into sub-curves:
        sectionsPerCurve = int(spt / cpt)

        # Create a list of point tags that will split the spline:
        # Basically, they are the points to divide the spirals into sectionsPerCurve
        # turns. However, some other points are also included to support the float
        # number of turns. It is best to visually look at the divisions with
        # gmsh.fltk.run() to understand why the split points are chosen the way they are
        # selected.
        if cutPlaneNormal is None:
            pointObjectsWithoutBreakPoints = points
            pointTagsWithoutBreakPoints = pointTags

        splitPointTags = list(
            set(pointTagsWithoutBreakPoints[:-1:sectionsPerCurve])
            | set(pointTagsWithoutBreakPoints[-spt - 1 :: -spt])
            | set(breakPointTagsDueToCutPlane)
            | set(breakPointTagsDueToTransition)
        )
        splitPointTags = sorted(splitPointTags)
        # Remove the first element of the list (starting point):
        _, *splitPointTags = splitPointTags

        # Also create a list of corresponding point objects:
        splitPoints = list(
            set(pointObjectsWithoutBreakPoints[:-1:sectionsPerCurve])
            | set(pointObjectsWithoutBreakPoints[-spt - 1 :: -spt])
            | set(breakPointObjectsDueToCutPlane)
            | set(breakPointObjectsDueToTransition)
        )
        splitPoints = sorted(splitPoints, key=lambda x: x.tag)
        # Remove the first element of the list (starting point):
        _, *splitPoints = splitPoints

        # Split the spline:
        dims = [0] * len(splitPointTags)
        _, splines = gmsh.model.occ.fragment(
            [(1, spline)],
            list(zip(dims, splitPointTags)),
            removeObject=True,
            removeTool=True,
        )
        splines = splines[0]
        self.splineTags = [j for _, j in splines]

        # Note the turn number of each spline. This will be used in getSplineTag and
        # getSplineTags methods.
        self.splineTurns = []
        for i in range(len(self.splineTags)):
            if i == 0:
                startPoint = points[0]
                endPoint = splitPoints[0]
            elif i == len(self.splineTags) - 1:
                startPoint = splitPoints[-1]
                endPoint = points[-1]
            else:
                startPoint = splitPoints[i - 1]
                endPoint = splitPoints[i]

            startTurn = (startPoint.theta - initialTheta) / (2 * math.pi)
            startTurn = round(startTurn / self.turnRes) * self.turnRes
            endTurn = (endPoint.theta - initialTheta) / (2 * math.pi)
            endTurn = round(endTurn / self.turnRes) * self.turnRes

            if direction is direction.ccw:
                self.splineTurns.append((startTurn, endTurn))
            else:
                self.splineTurns.append((-startTurn, -endTurn))

        # Check if splineTurn tuples starts with the small turn number:
        for i in range(len(self.splineTurns)):
            self.splineTurns[i] = sorted(self.splineTurns[i])

        # =============================================================================
        # GENERATING SPLINES ENDS =====================================================
        # =============================================================================

        # Find start and end points of the spiral curve:
        gmsh.model.occ.synchronize()  # synchronize the model to make getBoundary work
        self.startPointTag = gmsh.model.getBoundary([(1, self.getSplineTag(0))])[1][1]
        self.endPointTag = gmsh.model.getBoundary(
            [(1, self.getSplineTag(self.turns, endPoint=True))]
        )[1][1]

    def getSplineTag(self, turn, endPoint=False):
        """
        Returns the spline tag at a specific turn. It returns the spline tag of the
        section that is on the turn except its end point.

        :param turn: turn number (it can be a float)
        :type turn: float
        :param endPoint: if True, return the spline tag of the section that is on the
            turn including its end point but not its start point (default: False)
        :type endPoint: bool, optional
        :return: spline tag
        """
        if endPoint:
            for i, r in enumerate(self.splineTurns):
                if r[0] + (self.turnRes / 2) < turn <= r[1] + (self.turnRes / 2):
                    return self.splineTags[i]
        else:
            for i, r in enumerate(self.splineTurns):
                if r[0] - (self.turnRes / 2) <= turn < r[1] - (self.turnRes / 2):
                    return self.splineTags[i]

    def getPointTag(self, turn, endPoint=False):
        """
        Returns the point object at a specific turn.

        :param turn: turn number (it can be a float)
        :type turn: float
        :return: point object
        :rtype: point
        """
        if turn < 0 or turn > self.turns:
            raise ValueError("Turn number is out of range!")

        if turn == 0:
            return self.startPointTag
        elif turn == self.turns:
            return self.endPointTag
        else:
            curveTag = self.getSplineTag(turn, endPoint=endPoint)
            if endPoint:
                points = gmsh.model.getBoundary([(1, curveTag)])
                return points[1][1]
            else:
                points = gmsh.model.getBoundary([(1, curveTag)])
                return points[0][1]

    def getSplineTags(self, turnStart=None, turnEnd=None):
        """
        Get the spline tags from a specific turn to another specific turn. If turnStart
        and turnEnd are not specified, it returns all the spline tags.

        :param turnStart: start turn number (it can be a float) (default: None)
        :type turnStart: float, optional
        :param turnEnd: end turn number (it can be a float) (default: None)
        :type turnEnd: float, optional
        :return: spline tags
        :rtype: list[int]
        """
        if turnStart is None and turnEnd is None:
            return self.splineTags
        elif turnStart is None or turnEnd is None:
            raise ValueError(
                "turnStart and turnEnd must be both specified or both not specified."
                " You specified only one of them."
            )
        else:
            start = self.splineTags.index(self.getSplineTag(turnStart, False))
            end = self.splineTags.index(self.getSplineTag(turnEnd, True)) + 1
            return self.splineTags[start:end]




import os
from pathlib import Path
import gmsh
import json
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.lines as lines
import matplotlib.patches as patches
from matplotlib.collections import PatchCollection

from fiqus.utils.Utils import GmshUtils
from fiqus.utils.Utils import GeometricFunctions as Func
from fiqus.utils.Utils import RoxieParsers
from fiqus.utils.Utils import FilesAndFolders as Util
from fiqus.data import DataFiQuS as dF
from fiqus.data import DataMultipole as dM
from fiqus.data import RegionsModelFiQuS as rM


class PostProcess:
    def __init__(self, data: dF.FDM() = None, solution_folder: str = None, verbose: bool = False):
        """
        Class to post process results
        :param data: FiQuS data model
        :param verbose: If True more information is printed in python console.
        """
        self.data: dF.FDM() = data
        self.solution_folder = solution_folder
        self.verbose: bool = verbose
        self.md = dM.MultipoleData()
        self.rm = rM.RegionsModel()

        self.gu = GmshUtils(self.solution_folder, self.verbose)
        self.gu.initialize(verbosity_Gmsh=self.data.run.verbosity_Gmsh)

        self.brep_iron_curves = {1: set(), 2: set(), 3: set(), 4: set()}
        self.strands = None
        self.crns = None
        self.avg_temperatures = pd.DataFrame()
        self.postprocess_parameters = dict.fromkeys(['overall_error', 'minimum_diff', 'maximum_diff'])
        self.mesh_folder = os.path.dirname(self.solution_folder)
        self.geom_files = os.path.join(os.path.dirname(self.mesh_folder), self.data.general.magnet_name)
        self.mesh_files = os.path.join(self.mesh_folder, self.data.general.magnet_name)
        self.model_file = os.path.join(self.solution_folder, self.data.general.magnet_name)
        self.postproc_settings = pd.DataFrame()

        self.supported_variables = {'magnetic_flux_density': 'b',
                                    'temperature': 'T'}
        if any([var not in self.supported_variables.values()
                for var in self.data.magnet.postproc.electromagnetics.variables + self.data.magnet.postproc.thermal.variables]):
            pass
            # raise Exception(f"The interpolation of the field at the strands locations can not be executed: "
            #                 f"a variable listed in 'post_processors' -> 'variables' is not supported. "
            #                 f"Supported variables are: {self.supported_variables.values()}")
        self.physical_quantities_abbreviations = \
            {'magnetic_flux_density': ('BX/T', 'BY/T'),
             'temperature':           ('T/K', '-')}
        self.physical_quantity = None
        self.formatted_headline = "{0:>5}{1:>8}{2:>7}{3:>12}{4:>13}{5:>8}{6:>11}{7:>16}{8:>8}{9:>10}\n\n"
        self.formatted_content = "{0:>6}{1:>6}{2:>7}{3:>13}{4:>13}{5:>11}{6:>11}{7:>11}{8:>9}{9:>8}\n"
        self.map2d_headline_names = []

    def prepare_settings(self, settings):
        self.postproc_settings = pd.DataFrame({
            'variables': settings.variables,
            'volumes': settings.volumes})
        if 'compare_to_ROXIE' in settings.dict():
            self.physical_quantity = 'magnetic_flux_density'
        else:
            self.physical_quantity = 'temperature'
        self.map2d_headline_names = ['BL.', 'COND.', 'NO.', 'X-POS/MM', 'Y-POS/MM'] + \
                                    [abbr for abbr in self.physical_quantities_abbreviations[self.physical_quantity]] + \
                                    ['AREA/MM**2', 'CURRENT', 'FILL FAC.']

        if settings.plot_all != 'false':
            self.fiqus = None
            self.roxie = None
            fig1 = plt.figure(1)
            self.ax = fig1.add_subplot()
            self.ax.set_xlabel('x [cm]')  # adjust other plots to cm
            self.ax.set_ylabel('y [cm]')
            # self.ax.set_xlim(0, 0.09)  # todo
            # self.ax.set_ylim(0, 0.09)

            if not settings.dict().get('take_average_conductor_temperature', False):
                if settings.dict().get('compare_to_ROXIE', False):
                    fig2 = plt.figure(2)
                    self.ax2 = fig2.add_subplot(projection='3d')
                    self.ax2.set_xlabel('x [m]')
                    self.ax2.set_ylabel('y [m]')
                    self.ax2.set_zlabel('Absolute Error [T]')
                    self.fig4 = plt.figure(4)
                    self.ax4 = plt.axes()
                    self.ax4.set_xlabel('x [cm]')
                    self.ax4.set_ylabel('y [cm]')
                    self.ax4.set_aspect('equal', 'box')
                fig3 = plt.figure(3)
                self.ax3 = fig3.add_subplot(projection='3d')
                self.ax3.set_xlabel('x [m]')
                self.ax3.set_ylabel('y [m]')
                self.ax3.set_zlabel('norm(B) [T]' if 'compare_to_ROXIE' in settings.dict() else '')

        if 'compare_to_ROXIE' in settings.dict() and 'b' in settings.variables:
            b_index = settings.variables.index('b')
            file_to_open = os.path.join(self.solution_folder, f"b_{settings.volumes[b_index]}.pos")
        elif 'T' in settings.variables:
            T_index = settings.variables.index('T')
            file_to_open = os.path.join(self.solution_folder, f"T_{settings.volumes[T_index]}.pos")
            
        gmsh.open(file_to_open)

    def clear(self):
        self.md = dM.MultipoleData()
        self.rm = rM.RegionsModel()
        plt.close('all')
        gmsh.clear()

    def ending_step(self, gui: bool = False):
        if gui:
            self.gu.launch_interactive_GUI()
        else:
            gmsh.clear()
            gmsh.finalize()

    def loadAuxiliaryFile(self, run_type):
        self.md = Util.read_data_from_yaml(f"{self.mesh_files}_{run_type}.aux", dM.MultipoleData)

    def loadRegionFile(self):
        self.rm = Util.read_data_from_yaml(f"{self.mesh_files}_TH.reg", rM.RegionsModel)

    def loadStrandPositions(self, run_type):
        self.strands = json.load(open(f"{self.geom_files}_{run_type}.strs"))

    def loadHalfTurnCornerPositions(self):
        self.crns = json.load(open(f"{self.geom_files}.crns"))

    def plotHalfTurnGeometry(self, compare_to_ROXIE):
        for i in range(len(self.crns['iH'])):
            self.ax.add_line(lines.Line2D([self.crns['iH'][i][0], self.crns['iL'][i][0]],
                                          [self.crns['iH'][i][1], self.crns['iL'][i][1]], color='green'))
            self.ax.add_line(lines.Line2D([self.crns['oH'][i][0], self.crns['oL'][i][0]],
                                          [self.crns['oH'][i][1], self.crns['oL'][i][1]], color='green'))
            self.ax.add_line(lines.Line2D([self.crns['oL'][i][0], self.crns['iL'][i][0]],
                                          [self.crns['oL'][i][1], self.crns['iL'][i][1]], color='green'))
            self.ax.add_line(lines.Line2D([self.crns['iH'][i][0], self.crns['oH'][i][0]],
                                          [self.crns['iH'][i][1], self.crns['oH'][i][1]], color='green'))
            cc_fiqus = Func.centroid([self.crns['iH'][i][0], self.crns['iL'][i][0],
                                      self.crns['oL'][i][0], self.crns['oH'][i][0]],
                                     [self.crns['iH'][i][1], self.crns['iL'][i][1],
                                      self.crns['oL'][i][1], self.crns['oH'][i][1]])

            if compare_to_ROXIE:
                self.ax.add_line(lines.Line2D([self.crns['iHr'][i][0], self.crns['iLr'][i][0]],
                                              [self.crns['iHr'][i][1], self.crns['iLr'][i][1]],
                                              color='red', linestyle='dashed'))
                self.ax.add_line(lines.Line2D([self.crns['oHr'][i][0], self.crns['oLr'][i][0]],
                                              [self.crns['oHr'][i][1], self.crns['oLr'][i][1]],
                                              color='red', linestyle='dashed'))
                self.ax.add_line(lines.Line2D([self.crns['oLr'][i][0], self.crns['iLr'][i][0]],
                                              [self.crns['oLr'][i][1], self.crns['iLr'][i][1]],
                                              color='red', linestyle='dashed'))
                self.ax.add_line(lines.Line2D([self.crns['iHr'][i][0], self.crns['oHr'][i][0]],
                                              [self.crns['iHr'][i][1], self.crns['oHr'][i][1]],
                                              color='red', linestyle='dashed'))
                self.ax.text((self.crns['oLr'][i][0] + self.crns['iLr'][i][0]) / 2,
                             (self.crns['oLr'][i][1] + self.crns['iLr'][i][1]) / 2,
                             'R' + str(i + 1), style='italic', bbox={'facecolor': 'red', 'pad': 2})
                self.ax.text((self.crns['iHr'][i][0] + self.crns['oHr'][i][0]) / 2,
                             (self.crns['iHr'][i][1] + self.crns['oHr'][i][1]) / 2,
                             'L' + str(i + 1), style='italic', bbox={'facecolor': 'red', 'pad': 2})
                cc_roxie = Func.centroid(
                    [self.crns['iHr'][i][0], self.crns['iLr'][i][0], self.crns['oLr'][i][0], self.crns['oHr'][i][0]],
                    [self.crns['iHr'][i][1], self.crns['iLr'][i][1], self.crns['oLr'][i][1], self.crns['oHr'][i][1]])
                self.roxie = self.ax.scatter(cc_roxie[0], cc_roxie[1], edgecolor='r', facecolor='none')

            self.fiqus = self.ax.scatter(cc_fiqus[0], cc_fiqus[1], c="green")

    def postProcess(self, postproc):
        df_ref = pd.DataFrame()
        model_file_extension = 'EM' if 'compare_to_ROXIE' in postproc.dict() else 'TH'

        if postproc.dict().get('compare_to_ROXIE', False):
            # flag_self_field = False
            # path_map2d = Path(postproc.compare_to_ROXIE, "MQXA_All_" +
            #                   f"{'WithIron_' if self.data.magnet.geometry.with_iron_yoke else 'NoIron_'}" +
            #                   f"{'WithSelfField' if flag_self_field else 'NoSelfField'}" +
            #                   f"{'' if flag_contraction else '_no_contraction'}" + ".map2d")
            df_ref = RoxieParsers.parseMap2d(map2dFile=Path(postproc.compare_to_ROXIE), physical_quantity='magnetic_flux_density')
            BB_roxie = np.linalg.norm(df_ref[['BX/T', 'BY/T']].values, axis=1)
            if postproc.plot_all != 'false':
                path_cond2d = Path(os.path.join(os.path.dirname(postproc.compare_to_ROXIE), self.data.general.magnet_name + ".cond2d"))
                # path_cond2d = Path(os.path.dirname(postproc.compare_to_ROXIE), "MQXA_All_NoIron_NoSelfField" +
                #                    f"{'' if flag_contraction else '_no_contraction'}" + ".cond2d")
                if os.path.isfile(path_cond2d):
                    conductorPositionsList = RoxieParsers.parseCond2d(path_cond2d)

        # Collect strands coordinates
        strands_x = df_ref['X-POS/MM'] / 1e3 if postproc.dict().get('compare_to_ROXIE', False) else self.strands['x']
        strands_y = df_ref['Y-POS/MM'] / 1e3 if postproc.dict().get('compare_to_ROXIE', False) else self.strands['y']

        # Probe physical quantity values from view and region areas
        physical_quantity_values = {'x': [], 'y': []}
        cond_areas, current_signs = [], []
        if postproc.dict().get('take_average_conductor_temperature', False):
            half_turns = {name[:-3]: str(values.vol.numbers[i])
                          for group, values in self.rm.powered.items() for i, name in enumerate(values.vol.names)}
            self.avg_temperatures = pd.concat([pd.read_csv(os.path.join(self.solution_folder, 'T_avg', 'T_avg_0.txt'),
                                                           delimiter=r'\s+', header=None, usecols=[0], names=['Time'])] +
                                              [pd.read_csv(os.path.join(self.solution_folder, 'T_avg', f'T_avg_{i}.txt'),
                                                           delimiter=r'\s+', header=None, usecols=[1], names=[ht.upper()]) for i, ht in enumerate(half_turns)], axis=1)
            self.avg_temperatures['Time'] = pd.read_csv(os.path.join(self.solution_folder, 'T_avg', 'T_avg_0.txt'),
                                                        delimiter=r'\s+', header=None, usecols=[0], names=['Time'])['Time']
            self.avg_temperatures = self.avg_temperatures[['Time'] + ['HT' + str(i) for i in range(1, self.strands['ht'][-1] + 1)]]
            columns_to_format = self.avg_temperatures.columns[1:]
            self.avg_temperatures[columns_to_format] = self.avg_temperatures[columns_to_format].round(4)
            self.avg_temperatures.to_csv(os.path.join(self.solution_folder, 'half_turn_temperatures_over_time.csv'), index=False)
        else:
            print(f"Info    : {self.data.general.magnet_name} - I n t e r p o l a t i n g . . .")
            print(f"Info    : Interpolating {'magnetic flux density' if 'compare_to_ROXIE' in postproc.dict() else 'temperature'} ...")

            # view = gmsh.view.getTags()[0] if len(postproc.variables) == 1 else self.postproc_settings[
            #     (self.postproc_settings['variables'] == self.supported_variables[self.physical_quantity]) &
            #     (self.postproc_settings['volumes'] == 'Omega_p')].index[0]

            view = gmsh.view.getTags()[0]
            for i in range(len(strands_x)):
                is_new_conductor = i == 0 or self.strands['ht'][i] != self.strands['ht'][i - 1]
                is_new_block = i == 0 or self.strands['block'][i] != self.strands['block'][i - 1]

                # Print update
                if is_new_block:
                    perc = round(self.strands['block'][i] / self.strands['block'][-1] * 100)
                    print(f"Info    : [{' ' if perc < 10 else ''}{' ' if perc < 100 else ''}{perc}%] Interpolating within block {self.strands['block'][i]}")

                # Probe
                probe_data = gmsh.view.probe(view, strands_x[i], strands_y[i], 0)[0]
                if 'compare_to_ROXIE' in postproc.dict():
                    physical_quantity_values['x'].append(probe_data[0])
                    physical_quantity_values['y'].append(probe_data[1])
                else:
                    physical_quantity_values['x'].append(probe_data[0])
                    physical_quantity_values['y'].append(0)

                # Plot conductor and block identifiers
                if postproc.dict().get('compare_to_ROXIE', False) and postproc.plot_all != 'false':
                    if is_new_conductor:
                        self.ax.text(df_ref['X-POS/MM'][i] / 1e3, df_ref['Y-POS/MM'][i] / 1e3, str(self.strands['ht'][i]),
                                     style='italic', bbox={'facecolor': 'blue', 'pad': 3})
                    if is_new_block:
                        mid_strand_index = round(self.strands['ht'].count(self.strands['ht'][i]) / 2)
                        self.ax.text(df_ref['X-POS/MM'][i + mid_strand_index] / 1e3, df_ref['Y-POS/MM'][i + mid_strand_index] / 1e3,
                                     str(self.strands['block'][i]), style='italic', bbox={'facecolor': 'green', 'pad': 3})

                # Get current sign
                current_signs.append(self.md.domains.physical_groups.blocks[self.strands['block'][i]].current_sign)

                # Get region area
                if is_new_conductor:
                    gmsh.plugin.setNumber("MeshVolume", "Dimension", 2)
                    gmsh.plugin.setNumber("MeshVolume", "PhysicalGroup",
                                          self.md.domains.physical_groups.blocks[self.strands['block'][i]].half_turns[self.strands['ht'][i]].tag)
                    gmsh.plugin.run("MeshVolume")
                    cond_areas.append(gmsh.view.getListData(gmsh.view.getTags()[-1])[2][-1][-1])
                else:
                    cond_areas.append(cond_areas[-1])

            print(f"Info    : {self.data.general.magnet_name} - E n d   I n t e r p o l a t i n g")

            # Assemble map2d content
            strands_nr = 0
            content = []
            for i, ht in enumerate(self.strands['ht']):
                if i == 0 or ht != self.strands['ht'][i - 1]:
                    strands_nr = self.strands['ht'].count(ht)
                content.append(self.formatted_content.format(
                    int(self.strands['block'][i]),                  # bl
                    int(ht),                                        # cond
                    int(i + 1),                                     # no
                    f"{strands_x[i] * 1e3:.4f}",                    # x
                    f"{strands_y[i] * 1e3:.4f}",                    # y
                    f"{physical_quantity_values['x'][i]:.4f}",      # pq_x
                    f"{physical_quantity_values['y'][i]:.4f}",      # pq_y
                    f"{cond_areas[i] / strands_nr * 1e6:.4f}",      # area
                    f"{current_signs[i] * self.data.power_supply.I_initial / strands_nr:.2f}",                  # curr
                    f"{df_ref['FILL FAC.'][i] if postproc.dict().get('compare_to_ROXIE', False) else 0:.4f}"))    # fill_fac

            # Save map2d file
            with open(f"{self.model_file}_{model_file_extension}.map2d", 'w') as file:
                file.write(self.formatted_headline.format(*self.map2d_headline_names))
                file.writelines(content)
            print(f"Info    : Map2D file saved.")
            print(f"WARNING : [Map2D] All strand surface areas are equal within a conductor. Refer to the ROXIE map2d file for actual values")
            if not postproc.dict().get('compare_to_ROXIE', True):
                print(f"WARNING : [Map2D] No data is available for Filling Factor. Refer to the ROXIE map2d file for correct values")

        # Compute errors
        pq = np.linalg.norm(np.column_stack((np.array(physical_quantity_values['x']), np.array(physical_quantity_values['y']))), axis=1)
        if postproc.dict().get('compare_to_ROXIE', False):
            BB_err = pq - BB_roxie
            self.postprocess_parameters['overall_error'] = np.mean(abs(BB_err))
            self.postprocess_parameters['minimum_diff'] = np.min(BB_err)
            self.postprocess_parameters['maximum_diff'] = np.max(BB_err)

        if postproc.plot_all != 'false':
            if postproc.dict().get('take_average_conductor_temperature', False):
                min_value = self.avg_temperatures.iloc[:, 1:].min().min()
                max_value = self.avg_temperatures.iloc[:, 1:].max().max()
                ht_polygons = [patches.Polygon(np.array([(self.crns['iHr'][i][0], self.crns['iHr'][i][1]),
                                                         (self.crns['iLr'][i][0], self.crns['iLr'][i][1]),
                                                         (self.crns['oLr'][i][0], self.crns['oLr'][i][1]),
                                                         (self.crns['oHr'][i][0], self.crns['oHr'][i][1])]) * 1e2,
                                               closed=True) for i in range(len(self.crns['iHr']))]
                collection = PatchCollection(ht_polygons)
                self.ax.add_collection(collection)
                cmap = plt.get_cmap('plasma')
                norm = plt.Normalize(vmin=min_value, vmax=max_value)
                cbar = plt.colorbar(plt.cm.ScalarMappable(cmap=cmap, norm=norm), ax=self.ax)
                cbar.set_label('Temperature [K]')
                self.ax.autoscale_view()
                for i in range(self.avg_temperatures['Time'].size):
                    collection.set_facecolor(cmap(norm(self.avg_temperatures.iloc[i, 1:])))
                    if postproc.plot_all == 'true':
                        plt.pause(0.05)

            else:
                self.plotHalfTurnGeometry(postproc.dict().get('compare_to_ROXIE', False))
                map2d_strands = self.ax.scatter(strands_x, strands_y, edgecolor='black', facecolor='black', s=10)

                scatter3D_pos = self.ax3.scatter3D(strands_x, strands_y, pq, c=pq, cmap='Greens', vmin=0, vmax=10)

                if postproc.dict().get('compare_to_ROXIE', False):
                    if os.path.isfile(path_cond2d):
                        conductors_corners = [condPos.xyCorner for condPos in conductorPositionsList]
                        for corners in conductors_corners:
                            for corner in range(len(corners)):
                                self.ax.scatter(corners[corner][0] / 1e3, corners[corner][1] / 1e3, edgecolor='black', facecolor='black', s=10)

                    self.ax2.scatter3D(strands_x, strands_y, BB_err, c=BB_err, cmap='viridis')  # , vmin=-0.2, vmax=0.2)
                    scatter4 = self.ax4.scatter(np.array(strands_x) * 1e2, np.array(strands_y) * 1e2, s=1, c=np.array(BB_err) * 1e3, cmap='viridis')
                    scatter3D_pos_roxie = self.ax3.scatter3D(strands_x, strands_y, BB_roxie, c=BB_roxie, cmap='Reds', vmin=0, vmax=10)

                    cax4 = self.fig4.add_axes((self.ax4.get_position().x1 + 0.02, self.ax4.get_position().y0,
                                               0.02, self.ax4.get_position().height))
                    cbar = plt.colorbar(scatter4, cax=cax4)
                    cbar.ax.set_ylabel('Absolute error [mT]', rotation=270)
                    self.ax3.legend([scatter3D_pos, scatter3D_pos_roxie], ['FiQuS', 'ROXIE'], numpoints=1)
                    self.ax.legend([self.fiqus, self.roxie, map2d_strands], ['FiQuS', 'ROXIE'], numpoints=1)
                    self.fig4.savefig(f"{os.path.join(self.solution_folder, self.data.general.magnet_name)}.svg", bbox_inches='tight')

            if postproc.plot_all == 'true':
                plt.show()

        # os.remove(os.path.join(self.solution_folder, 'b_Omega_p.pos'))
        # os.remove(f"{os.path.join(self.solution_folder, self.data.general.magnet_name)}.pre")
        # os.remove(f"{os.path.join(os.path.dirname(self.solution_folder), self.data.general.magnet_name)}.msh")

    def completeMap2d(self):
        def _quadrant(x, y):
            if x < 0 and y < 0: return 3
            elif x < 0: return 2
            elif y < 0: return 4
            else: return 1

        if self.data.magnet.geometry.electromagnetics.symmetry == 'xy':
            if self.strands['poles'] == 2:
                mirror_components = {2: [-1, 1], 3: [1, 1], 4: [-1, 1]}
            elif self.strands['poles'] == 4:
                mirror_components = {2: [1, -1], 3: [-1, -1], 4: [-1, 1]}
        elif self.data.magnet.geometry.electromagnetics.symmetry == 'x':
            if self.strands['poles'] == 2:
                mirror_components = {3: [-1, 1], 4: [-1, 1]}
            elif self.strands['poles'] == 4:
                mirror_components = {3: [-1, 1], 4: [-1, 1]}
        elif self.data.magnet.geometry.electromagnetics.symmetry == 'y':
            if self.strands['poles'] == 2:
                mirror_components = {2: [-1, 1], 3: [-1, 1]}
            elif self.strands['poles'] == 4:
                mirror_components = {2: [1, -1], 3: [1, -1]}
        else:
            mirror_components = {}

        print(f"Info    : {self.data.general.magnet_name} - M i r r o r i n g . . .")
        print(f"Info    : Mirroring by symmetry ...")
        blocks_nr = self.strands['block'][-1]

        with open(f"{self.model_file}_EM.map2d", 'r') as file:
            file_content = file.read()
            content_by_row = file_content.split('\n')
        new_content = [content_by_row[0] + '\n' + content_by_row[1] + '\n']
        prev_block: int = 0
        for row in content_by_row[2:-1]:
            entries = row.split()
            str_nr, x_coord, y_coord = entries[2], float(entries[3]), float(entries[4])
            qdr = _quadrant(x_coord, y_coord)
            if qdr in mirror_components:
                #found = re.search(f" {str(abs(x_coord))} +{str(abs(y_coord))}", file_content)
                #BB = [row_ref for row_ref in content_by_row if f" {abs(x_coord):.4f}       {abs(y_coord):.4f}" in row_ref][0].split()[5:7]
                BB = content_by_row[self.strands['mirrored'][str_nr] + 1].split()[5:7]
                row = row.replace(entries[5], f'{mirror_components[qdr][0] * float(BB[0]):.4f}')
                row = row.replace(entries[6], f'{mirror_components[qdr][1] * float(BB[1]):.4f}')
            if int(entries[0]) > prev_block:
                perc = round(int(entries[0]) / blocks_nr * 100)
                print("Info    : [" + f"{' ' if perc < 10 else ''}" + f"{' ' if perc < 100 else ''}" + f"{perc}" +
                      "%] Mirroring within block" + f"{entries[0]}")
                prev_block = int(entries[0])
            new_content.append(row + '\n')
        with open(f"{self.model_file}_EM.map2d", 'w') as file:
            file.writelines(new_content)
        print(f"Info    : {self.data.general.magnet_name} - E n d   M i r r o r i n g")



import os
import shutil

import fiqus.data.DataFiQuSPancake3D as Pancake3D
from fiqus.data.DataFiQuS import FDM
from fiqus.utils.Utils import FilesAndFolders as Util
from fiqus import MainFiQuS as mf


# Generate reference files for the models below:
model_names = [
    "TEST_Pancake3D_REF",
    "TEST_Pancake3D_REFStructured",
    "TEST_Pancake3D_TSA",
    "TEST_Pancake3D_TSAStructured",
    "TEST_Pancake3D_TSAInsulating"
]

for model_name in model_names:
    input_file = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        "_inputs",
        model_name,
        f"{model_name}.yaml",
    )
    # select _references folder as the output folder:
    output_folder = os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "_references", model_name
    )

    # Cast input yaml file to FDM
    data_model: FDM = Util.read_data_from_yaml(input_file, FDM)

    data_model.run.overwrite = True
    data_model.run.launch_gui = False

    # if the output folder exists, remove it and create a new one:
    if os.path.exists(output_folder):
        shutil.rmtree(output_folder)
        # Create the output folder:
        os.makedirs(output_folder)

    # Solve the same model three times with different solve types:
    solve_types = [
        "weaklyCoupled",
        "stronglyCoupled",
        "electromagnetic",
    ]
    for i, solve_type in enumerate(solve_types):
        if i == 0:
            # Make the run type start_from_yaml:
            data_model.run.type = "start_from_yaml"
        else:
            # Make the run type solve_only:
            data_model.run.type = "solve_only"

        # data_model.run.type = "post_process"
        # data_model.run.type = "solve_only"

        data_model.magnet.solve.type = solve_type
        data_model.run.solution = solve_type
        if solve_type in ["weaklyCoupled", "stronglyCoupled"]:
            data_model.magnet.solve.quantitiesToBeSaved = [
                Pancake3D.Pancake3DSolveSaveQuantity(
                    quantity="magneticField",
                ),
                Pancake3D.Pancake3DSolveSaveQuantity(
                    quantity="currentDensity",
                ),
                Pancake3D.Pancake3DSolveSaveQuantity(
                    quantity="temperature",
                ),
            ]
        elif solve_type == "electromagnetic":
            data_model.magnet.solve.quantitiesToBeSaved = [
                Pancake3D.Pancake3DSolveSaveQuantity(
                    quantity="magneticField",
                ),
                Pancake3D.Pancake3DSolveSaveQuantity(
                    quantity="currentDensity",
                ),
            ]

        fiqus_instance = mf.MainFiQuS(fdm=data_model, model_folder=output_folder)

        # remove fiqus_instance to avoid memory issues:
        del fiqus_instance

        # # remove logs directory and run_log.csv inside the output folder:
        # shutil.rmtree(os.path.join(output_folder, "logs"))
        # os.remove(os.path.join(output_folder, "run_log.csv"))



import os
import shutil
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../../'))) # Add the path to the fiqus package to the system path
from fiqus.data.DataFiQuS import FDM
from fiqus.utils.Utils import FilesAndFolders as Util
from fiqus import MainFiQuS as mf


# Generate reference files for the models below:
model_names = [
    "TEST_CAC_Strand_adaptiveMesh",
    "TEST_CAC_Strand_hexFilaments",
    "TEST_CAC_wireInChannel",
]
# The run types for the models above:
run_types = [
    'geometry_and_mesh', 
    'start_from_yaml', 
    'start_from_yaml',
]

for model_name, run_type in zip(model_names, run_types):
    # get path to the input file:
    input_file = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        "_inputs",
        model_name,
        f"{model_name}.yaml",
    )

    # select _references folder as the output folder:
    output_folder = os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "_references", model_name
    )

    # if the output folder exists, remove it:
    if os.path.exists(output_folder):
        shutil.rmtree(output_folder)

    # Create the output folder:
    os.makedirs(output_folder)

    # Cast input yaml file to FDM
    data_model: FDM = Util.read_data_from_yaml(input_file, FDM)

    data_model.run.overwrite = True

    # Make the run type start_from_yaml:
    data_model.run.type = run_type

    fiqus_instance = mf.MainFiQuS(
        fdm=data_model, model_folder=output_folder, input_file_path=input_file
    )

    # remove fiqus_instance to avoid memory issues:
    del fiqus_instance
